# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2016 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#

from lib389.paths import Paths

# Test that we can retrieve the settings from the paths object
def test_paths():
    # Make the paths object.
    p = Paths()
    # Get a value!
    v = p.version

# Test that if we make the path object, and we don't read a path from it
# the filecache state is False
def test_path_noread():
    p = Paths()
    assert(p._defaults_cached is False)
    p._read_defaults()
    assert(p._defaults_cached is True)

def test_path_exception():
    # Trigger the internal path find with a "bad location" and
    # make sure that we get the exception
    p = Paths()
    try:
        p._get_defaults_loc(search_paths=[])
        assert(False)
    except IOError:
        assert(True)




# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2016-2017 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import pytest

from lib389.tests.cli import topology as default_topology
from lib389.cli_base import LogCapture, FakeArgs
from lib389.plugins import ReferentialIntegrityPlugin
from lib389.cli_conf.plugins import referint as referint_cli


@pytest.fixture(scope="module")
def topology(request):
    topology = default_topology(request)

    plugin = ReferentialIntegrityPlugin(topology.standalone)
    if not plugin.exists():
        plugin.create()

    # we need to restart the server after enabling the plugin
    plugin.enable()
    topology.standalone.restart()
    topology.logcap.flush()

    return topology


def test_set_update_delay(topology):
    args = FakeArgs()

    args.value = 60
    referint_cli.manage_update_delay(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains('referint-update-delay set to "60"')
    topology.logcap.flush()

    args.value = None
    referint_cli.manage_update_delay(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("referint-update-delay: 60")
    topology.logcap.flush()

    args.value = 0
    referint_cli.manage_update_delay(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains('referint-update-delay set to "0"')
    topology.logcap.flush()

    args.value = None
    referint_cli.manage_update_delay(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("referint-update-delay: 0")
    topology.logcap.flush()

def test_add_membership_attr(topology):
    args = FakeArgs()

    args.value = "member2"
    referint_cli.add_membership_attr(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("successfully added membership attribute")
    topology.logcap.flush()

    referint_cli.display_membership_attr(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains(": member2")
    topology.logcap.flush()

def test_add_membership_attr_with_value_that_already_exists(topology):
    plugin = ReferentialIntegrityPlugin(topology.standalone)
    # setup test
    if not "uniqueMember" in plugin.get_membership_attr():
        plugin.add_membership_attr("uniqueMember")

    args = FakeArgs()

    args.value = "uniqueMember"
    referint_cli.add_membership_attr(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("already exists")
    topology.logcap.flush()

def test_remove_membership_attr_with_value_that_exists(topology):
    plugin = ReferentialIntegrityPlugin(topology.standalone)
    # setup test
    if not "uniqueMember" in plugin.get_membership_attr():
        plugin.add_membership_attr("uniqueMember")

    args = FakeArgs()

    args.value = "uniqueMember"
    referint_cli.remove_membership_attr(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("successfully removed membership attribute")
    topology.logcap.flush()

    referint_cli.display_membership_attr(topology.standalone, None, topology.logcap.log, args)
    assert not topology.logcap.contains(": uniqueMember")
    topology.logcap.flush()

def test_remove_membership_attr_with_value_that_doesnt_exist(topology):
    args = FakeArgs()

    args.value = "whatever"
    referint_cli.remove_membership_attr(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains('No value "{0}" found'.format(args.value))
    topology.logcap.flush()

def test_try_remove_all_membership_attr_values(topology):
    plugin = ReferentialIntegrityPlugin(topology.standalone)
    #setup test
    membership_values = plugin.get_membership_attr()
    assert len(membership_values) > 0
    for val in membership_values[:-1]:
        plugin.remove_membership_attr(val)

    args = FakeArgs()

    args.value = membership_values[-1]
    referint_cli.remove_membership_attr(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("Error: Failed to delete. At least one value for membership attribute should exist.")
    topology.logcap.flush()



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2016, William Brown <william at blackhats.net.au>
# Copyright (C) 2024 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

from lib389._mapped_object import DSLdapObjects
from lib389.idm.account import Account

from lib389.utils import ds_is_older

RDN = 'cn'
MUST_ATTRIBUTES = [
    'cn',
]
DEFAULT_BASEDN_RDN = 'ou=Services'

class ServiceAccount(Account):
    """A single instance of Service entry

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """
    _must_attributes = [
        'cn',
        'description',
    ]

    def __init__(self, instance, dn=None):
        super(ServiceAccount, self).__init__(instance, dn)
        self._rdn_attribute = RDN
        self._must_attributes = MUST_ATTRIBUTES
        self._create_objectclasses = [
            'top',
            'applicationProcess',
        ]
        if ds_is_older('1.4.0', instance=instance):
            # This is a HORRIBLE HACK for older versions that DON'T have
            # correct updated schema!
            #
            # I feel physically ill having wrtten this line of code. :(
            self._create_objectclasses.append('extensibleobject')
        else:
            self._create_objectclasses.append('nsMemberOf')
            self._create_objectclasses.append('nsAccount')
        self._protected = False

class ServiceAccounts(DSLdapObjects):
    """DSLdapObjects that represents Services entry
    By default it uses 'ou=Services' as rdn.

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param basedn: Base DN for all group entries below
    :type basedn: str
    """

    def __init__(self, instance, basedn, rdn=DEFAULT_BASEDN_RDN):
        super(ServiceAccounts, self).__init__(instance)
        self._objectclasses = [
            'applicationProcess',
        ]
        self._filterattrs = [RDN]
        self._childobject = ServiceAccount

        if rdn is None:
            self._basedn = basedn
        else:
            self._basedn = f'{rdn},{basedn}'


    def create_test_service(self, description="Test Service"):
        """Create a test service with cn=test_service rdn

        :param description: Service description
        :type description: str

        :returns: DSLdapObject of the created entry
        """

        rdn_value = "test_service"
        rdn = "cn={}".format(rdn_value)
        properties = {
            'cn': rdn_value,
            'description': description,
        }
        return super(ServiceAccounts, self).create(rdn, properties)




#!/usr/bin/python3

# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2021 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import re
import sys
import os, os.path

# regex that matches a BIND request line
regex_num = r'[-]?\d+' # matches numbers including negative
regex_new_conn = re.compile(r'^(\[.+\]) (conn=%s) (fd=%s) (slot=%s) (?:SSL )?connection from (\S+)' % (regex_num, regex_num, regex_num))
regex_sslinfo = re.compile(r'^\[.+\] (conn=%s) SSL (.+)$' % regex_num)
regex_bind_req = re.compile(r'^(\[.+\]) (conn=%s) (op=%s) BIND dn=(.+) method=(\S+) version=\d ?(?:mech=(\S+))?' % (regex_num, regex_num))
regex_bind_res = re.compile(r'^(\[.+\]) (conn=%s) (op=%s) RESULT err=(%s) tag=97 ' % (regex_num, regex_num, regex_num))
regex_unbind = re.compile(r'^\[.+\] (conn=%s) op=%s UNBIND' % (regex_num, regex_num))
regex_closed = re.compile(r'^(\[.+\]) (conn=%s) (op=%s) fd=%s closed' % (regex_num, regex_num, regex_num))
regex_ssl_map_fail = re.compile(r'^\[.+\] (conn=%s) (SSL failed to map client certificate.*)$' % regex_num)

# bind errors we can ignore
ignore_errors = {'0': 'Success',
                 '10': 'Referral',
                 '14': 'SASL Bind In Progress'
                 }

REQ = 0
RES = 1

class Conn:
    def __init__(self, timestamp, conn, fd, slot, ip):
        self.conn = conn
        self.fd = fd
        self.slot = slot
        self.ip = ip
        self.timestamp = timestamp
        self.ops = {}
        self.sslinfo = ''

    def addssl(self, sslinfo):
        if self.sslinfo and sslinfo:
            self.sslinfo += ' '
        self.sslinfo += sslinfo

    def addreq(self, timestamp, opnum, dn, method, mech='SIMPLE'):
        retval = None
        if opnum in self.ops: # result came before request?
            op = self.ops.pop(opnum) # grab the op and remove from list
            if op[RES]['errnum'] in ignore_errors: # don't care about this op
                return retval
            if not mech: mech = "SIMPLE"
            op[REQ] = {'dn': dn, 'method': method, 'timestamp': timestamp,
                       'mech': mech}
            retval = self.logstr(opnum, op)
        else: # store request until we get the result
            op = [None, None] # new empty list
            if not mech: mech = "SIMPLE"
            op[REQ] = {'dn': dn, 'method': method, 'timestamp': timestamp,
                       'mech': mech}
            self.ops[opnum] = op
        return retval

    def addres(self, timestamp, opnum, errnum):
        retval = None
        if opnum in self.ops:
            op = self.ops.pop(opnum) # grab the op and remove from list
            if errnum in ignore_errors: # don't care about this op
                return retval
            op[RES] = {'errnum': errnum, 'timestamp': timestamp}
            retval = self.logstr(opnum, op)
        else: # result came before request in access log - store until we find request
            op = [None, None] # new empty list
            op[RES] = {'errnum': errnum, 'timestamp': timestamp}
            self.ops[opnum] = op
        return retval

    def logstr(self, opnum, op):
        # timestamp connnum opnum err=X request timestamp dn=Y method=Z mech=W timestamp ip=ip
        logstr = '%s %s %s err=%s REQUEST %s dn=%s method=%s mech=%s %s ip=%s extra=%s' % (
            op[RES]['timestamp'], self.conn, opnum, op[RES]['errnum'],
            op[REQ]['timestamp'], op[REQ]['dn'], op[REQ]['method'], op[REQ]['mech'],
            self.timestamp, self.ip, self.sslinfo
            )
        return logstr
        
# key is conn=X
# val is ops hash
#  key is op=Y
#  value is list
#    list[0] is BIND request
#    list[1] is RESULT
conns = {}

# file to log failed binds to
logf = None

def pre(plgargs):
    global logf
    logfile = plgargs.get('logfile', None)
    if not logfile:
        print("Error: missing required argument failedbinds.logfile")
        return False
    needchmod = False
    if not os.path.isfile(logfile): needchmod = True
    if sys.version_info < (3, 0):
        logf = open(logfile, 'a', 0) # 0 for unbuffered output
    else:
        logf = open(logfile, 'a')
    if needchmod: os.chmod(logfile, 0o600)
    return True

def post():
    global logf
    logf.close()
    logf = None

def plugin(line):
    # is this a new conn line?
    match = regex_new_conn.match(line)
    if match:
        (timestamp, connid, fdid, slotid, ip) = match.groups()
        if connid in conns: conns.pop(connid) # remove old one, if any
        conn = Conn(timestamp, connid, fdid, slotid, ip)
        conns[connid] = conn
        return True

    # is this an UNBIND line?
    match = regex_unbind.match(line)
    if match:
        connid = match.group(1)
        if connid in conns: conns.pop(connid) # remove it
        return True

    # is this a closed line?
    match = regex_closed.match(line)
    if match:
        (timestamp, connid, opid) = match.groups()
        if connid in conns: conns.pop(connid) # remove it
        return True

    # is this an SSL info line?
    match = regex_sslinfo.match(line)
    if match:
        (connid, sslinfo) = match.groups()
        if connid in conns:
            conns[connid].addssl(sslinfo)
        return True

    # is this a line with extra SSL mapping info?
    match = regex_ssl_map_fail.match(line)
    if match:
        (connid, sslinfo) = match.groups()
        if connid in conns:
            conns[connid].addssl(sslinfo)
        return True

    # is this a REQUEST line?
    match = regex_bind_req.match(line)
    if match:
        (timestamp, connid, opnum, dn, method, mech) = match.groups()
        # should have seen new conn line - if not, have to create a dummy one
        conn = conns.get(connid, Conn('unknown', connid, '', '', 'unknown'))
        logmsg = conn.addreq(timestamp, opnum, dn, method, mech)
        if logmsg:
            logf.write(logmsg + "\n")
            logf.flush()
        return True

    # is this a RESULT line?
    match = regex_bind_res.match(line)
    if match:
        (timestamp, connid, opnum, errnum) = match.groups()
        # should have seen new conn line - if not, have to create a dummy one
        conn = conns.get(connid, Conn('unknown', connid, '', '', 'unknown'))
        logmsg = conn.addres(timestamp, opnum, errnum)
        if logmsg:
            logf.write(logmsg + "\n")
            logf.flush()
        return True

    return True # no match



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2017 alisha17 <anejaalisha@yahoo.com>
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#
import pytest
import ldap

from lib389.tests.cli import topology as default_topology
from lib389.cli_base import LogCapture, FakeArgs
from lib389.plugins import AutoMembershipPlugin
from lib389.cli_conf.plugins import automember as automember_cli

@pytest.fixture(scope="module")
def topology(request):
    topology = default_topology(request)

    plugin = AutoMembershipPlugin(topology.standalone)
    if not plugin.exists():
        plugin.create()

    # we need to restart the server after enabling the plugin
    plugin.enable()
    topology.standalone.restart()
    topology.logcap.flush()

    return topology


def test_namenotexists_listdefinition(topology):
    """
        Test if non existent instance raises ldap.NO_SUCH_OBJECT exception 
        when showing automember definition.
    """
    args = FakeArgs()
    
    args.name = "somerandomcn"
    
    with pytest.raises(ldap.NO_SUCH_OBJECT):
        automember_cli.list_definition(topology.standalone, None, topology.logcap.log, args)
        

def test_createdefinition(topology):
    """
        Test if new automember definition is created successfully.
    """
    args = FakeArgs()

    args.name = "sometestcn"
    args.scope = "ou=People,dc=example,dc=com"
    args.filter = "objectclass=*"
    args.groupattr = "member:dn"
    args.defaultgroup = "cn=linux-group,cn=groups,dc=example,dc=com"
    
    automember_cli.create_definition(topology.standalone, None, topology.logcap.log, args)
    assert topology.logcap.contains("Automember definition created successfully!")

    topology.logcap.flush()


def test_invalidattributes_createdefinition(topology):
    """ 
        Test if invalid attributes raise ldap.INVALID_SYNTAX exception
        when creating new automember definition.
    """
    args = FakeArgs()

    args.name = "sometestcn"
    args.scope = "someinvalidscope"
    args.filter = "objectclass=*"
    args.groupattr = "member:dn"
    args.defaultgroup = "someinvaliddefaultgroup"

    with pytest.raises(ldap.INVALID_SYNTAX):
        automember_cli.create_definition(topology.standalone, None, topology.logcap.log, args)


def test_ifnameexists_createdefinition(topology):
    """
        Test if already existing instance name raises ldap.ALREADY-EXISTS
        exception when creating new automember definition.
    """

    args = FakeArgs()

    args.name = "sometestcn"
    args.scope = "ou=People,dc=example,dc=com"
    args.filter = "objectclass=*"
    args.groupattr = "member:dn"
    args.defaultgroup = "cn=windows-group,cn=groups,dc=example,dc=com"
    
    with pytest.raises(ldap.ALREADY_EXISTS):
        automember_cli.create_definition(topology.standalone, None, topology.logcap.log, args)


def test_editdefinition(topology):
    """
        Test if automember definition is updated successfully.
    """
    args = FakeArgs()

    args.name = "sometestcn"
    args.scope = "ou=People,dc=example,dc=com"
    args.filter = "objectclass=*"
    args.groupattr = "member:dn"
    args.defaultgroup = "cn=windows-group,cn=groups,dc=example,dc=com"

    a = automember_cli.edit_definition(topology.standalone, None, topology.logcap.log, args)

    assert topology.logcap.contains("Definition updated successfully.")

    
def test_nonexistentinstance_removedefinition(topology):
    """
        Test if non existent instance definition raises NO_SUCH_OBJECT
        definition when removing automember definition.
    """
    args = FakeArgs()

    args.name = "randomcn"

    with pytest.raises(ldap.NO_SUCH_OBJECT):
        automember_cli.remove_definition(topology.standalone, None, topology.logcap.log, args)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2018 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import os
import pytest
import shutil

from lib389.cli_conf.backup import backup_create, backup_restore
from lib389.cli_base import LogCapture, FakeArgs
from lib389.idm.user import UserAccounts
from lib389.tests.topologies import topology_st
from lib389.utils import ds_is_older
from lib389._constants import DEFAULT_SUFFIX
pytestmark = pytest.mark.skipif(ds_is_older('1.4.0'), reason="Not implemented")


def test_basic(topology_st):
    BACKUP_DIR = os.path.join(topology_st.standalone.ds_paths.backup_dir, "basic_backup")
    topology_st.logcap = LogCapture()
    args = FakeArgs()

    users = UserAccounts(topology_st.standalone, DEFAULT_SUFFIX)
    user = users.create_test_user()
    user.replace("description", "backup_test")

    # Clean the backup dir first
    if os.path.exists(BACKUP_DIR):
        shutil.rmtree(BACKUP_DIR)

    # Create the backup
    args.archive = BACKUP_DIR
    args.db_type = None
    backup_create(topology_st.standalone, None, topology_st.logcap.log, args)
    assert os.listdir(BACKUP_DIR)

    # Restore the backup
    args.archive = topology_st.standalone.ds_paths.backup_dir
    args.db_type = None
    backup_restore(topology_st.standalone, None, topology_st.logcap.log, args)
    assert user.present("description", "backup_test")

    # No error has happened! Done!
    # Clean up
    if os.path.exists(BACKUP_DIR):
        shutil.rmtree(BACKUP_DIR)



#!/usr/bin/python3

from __future__ import print_function

import sys
import os, os.path
import errno
import signal
import pprint
import types
import time
import fcntl
import pwd

maxlines = 1000 # set on command line
S_IFIFO = 0o010000

buffer = [] # default circular buffer used by default plugin
totallines = 0
logfname = "" # name of log pipe
debug = False

# default plugin just keeps a circular buffer
def defaultplugin(line):
    global totallines
    buffer.append(line)
    totallines = totallines + 1
    if len(buffer) > maxlines:
        del buffer[0]
    return True

def printbuffer():
    sys.stdout.writelines(buffer)
    print("Read %d total lines" % totallines)
    print(logfname, "=" * 60)
    sys.stdout.flush()

def defaultpost(): printbuffer()

plgfuncs = [] # list of plugin functions
plgpostfuncs = [] # list of post plugin funcs

def finish():
    for postfunc in plgpostfuncs: postfunc()
    if options.scriptpidfile: os.unlink(options.scriptpidfile)
    sys.exit(0)

def sighandler(signum, frame):
    if signum != signal.SIGHUP:
        signal.signal(signal.SIGHUP, signal.SIG_DFL)
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        #signal.signal(signal.SIGPIPE, signal.SIG_DFL)
        signal.signal(signal.SIGTERM, signal.SIG_DFL)
        signal.signal(signal.SIGALRM, signal.SIG_DFL)
        if signum == signal.SIGALRM and debug:
            print("script timed out waiting to open pipe")
        finish()
    else: printbuffer()

def isvalidpluginfile(plg):
    return os.path.isfile(plg)

def my_import(plgfile):
    '''import plgfile as a python module and return
    an error string if error - also return the prefunc if any'''
    if not isvalidpluginfile(plgfile):
        return ("%s is not a valid plugin filename" % plgfile, None, None)
    # __import__ searches for the file in sys.path - we cannot
    # __import__ a file by the full path
    # __import__('basename') looks for basename.py in sys.path
    (dir, fname) = os.path.split(plgfile)
    base = os.path.splitext(fname)[0]
    if not dir: dir = "."
    sys.path.insert(0, dir) # put our path first so it will find our file
    mod = __import__(base) # will throw exception if problem with python file
    sys.path.pop(0) # remove our path

    # check for the plugin functions
    plgfunc = getattr(mod, 'plugin', None)
    if not plgfunc:
        return ('%s does not specify a plugin function' % plgfile, None, base)
    if not isinstance(plgfunc, types.FunctionType):
        return ('the symbol "plugin" in %s is not a function' % plgfile, None, base)
    plgfuncs.append(plgfunc) # add to list in cmd line order

    # check for 'post' func
    plgpostfunc = getattr(mod, 'post', None)
    if plgpostfunc:
        if not isinstance(plgpostfunc, types.FunctionType):
            return ('the symbol "post" in %s is not a function' % plgfile, None, base)
        else:
            plgpostfuncs.append(plgpostfunc) # add to list in cmd line order

    prefunc = getattr(mod, 'pre', None)
    # check for 'pre' func
    if prefunc and not isinstance(prefunc, types.FunctionType):
        return ('the symbol "pre" in %s is not a function' % plgfile, None, base)

    return ('', prefunc, base)

def parse_plugins(parser, options, args):
    '''Each plugin in the plugins list may have additional
    arguments, specified on the command line like this:
    --plugin=foo.py foo.bar=1 foo.baz=2 ...
    that is, each argument to plugin X will be specified as X.arg=value'''
    if not options.plugins: return args

    for plgfile in options.plugins:
        (errstr, prefunc, base) = my_import(plgfile)
        if errstr:
            parser.error(errstr)
            return args

        # parse the arguments to the plugin given on the command line
        bvals = {} # holds plugin args and values, if any
        newargs = []
        for arg in args:
            if arg.startswith(base + '.'):
                argval = arg.replace(base + '.', '')
                (plgarg, plgval) = argval.split('=', 1) # split at first =
                if not plgarg in bvals:
                    bvals[plgarg] = plgval
                elif isinstance(bvals[plgarg],list):
                    bvals[plgarg].append(plgval)
                else: # convert to list
                    bvals[plgarg] = [bvals[plgarg], plgval]
            else:
                newargs.append(arg)
        if prefunc:
            if debug:
                print('Calling "pre" function in', plgfile)
            if not prefunc(bvals):
                parser.error('the "pre" function in %s returned an error' % plgfile)
        args = newargs

    return args

def open_pipe(logfname):
    opencompleted = False
    logf = None
    while not opencompleted:
        try:
            logf = open(logfname, 'r') # blocks until there is some input
            opencompleted = True
        except IOError as e:
            if e.errno == errno.EINTR:
                continue # open was interrupted, try again
            else: # hard error
                print("%s [%d]" % (e.strerror, e.errno))
                sys.exit(1)
    return logf

def is_proc_alive(procpid):
    retval = False
    try:
        retval = os.path.exists("/proc/%d" % procpid)
    except IOError as e:
        if e.errno != errno.ENOENT: # may not exist yet - that's ok
            # otherwise, probably permissions or other badness
            print("could not open file %s - %s [%d]" % (procfile, e.strerror, e.errno))
            sys.exit(1)
    # using /proc/pid failed, try kill
    if not retval:
        try:
            os.kill(procpid, 0) # sig 0 is a "ping"
            retval = True # if we got here, proc exists
        except OSError as e:
            pass # no such process, or EPERM/EACCES
    return retval

def get_pid_from_file(pidfile):
    procpid = 0
    if pidfile:
        line = None
        try:
            pfd = open(pidfile, 'r')
            line = pfd.readline()
            pfd.close()
        except IOError as e:
            if e.errno != errno.ENOENT: # may not exist yet - that's ok
                # otherwise, probably permissions or other badness
                print("Could not read pid from file %s - %s [%d]" % (pidfile, e.strerror, e.errno))
                sys.exit(1)
        if line:
            procpid = int(line)
    return procpid

def write_pid_file(pidfile):
    try:
        pfd = open(pidfile, 'w')
        pfd.write("%d\n" % os.getpid())
        pfd.close()
    except IOError as e:
        print("Could not write pid to file %s - %s [%d]" % (pidfile, e.strerror, e.errno))
        sys.exit(1)

def handle_script_pidfile(scriptpidfile):
    scriptpid = get_pid_from_file(scriptpidfile)
    # 0 if no file or no pid or error
    if scriptpid and is_proc_alive(scriptpid):
        # already running
        if debug:
            print("Script is already running: process id %d" % scriptpid)
        return False
    else:
        # either process is not running or no file
        # write our pid to the file
        write_pid_file(scriptpidfile)
    return True

def read_and_process_line(logf, plgfuncs):
    line = None
    done = False
    readcompleted = False
    while not readcompleted:
        try:
            line = logf.readline()
            readcompleted = True # read completed
        except IOError as e:
            if e.errno == errno.EINTR:
                continue # read was interrupted, try again
            else: # hard error
                print("%s [%d]" % (e.strerror, e.errno))
                sys.exit(1)
    if line: # read something
        for plgfunc in plgfuncs:
            if not plgfunc(line):
                print("Aborting processing due to function %s.%s" % (plgfunc.__module__, plgfunc.__name__))
                finish() # this will exit the process
                done = True
                break
    else: # EOF
        done = True
    return done

def parse_options():
    from optparse import OptionParser
    usage = "%prog <name of pipe> [options]"
    parser = OptionParser(usage)
    parser.add_option("-m", "--maxlines", dest="maxlines", type='int',
                      help="maximum number of lines to keep in the buffer", default=1000)
    parser.add_option("-d", "--debug", dest="debug", action="store_true",
                      default=False, help="gather extra debugging information")
    parser.add_option("-p", "--plugin", type='string', dest='plugins', action='append',
                      help='filename of a plugin to use with this log')
    parser.add_option("-s", "--serverpidfile", type='string', dest='serverpidfile',
                      help='name of file containing the pid of the server to monitor')
    parser.add_option("-t", "--servertimeout", dest="servertimeout", type='int',
                      help="timeout in seconds to wait for the serverpid to be alive. only applies when using -s or --serverpid", default=60)
    parser.add_option("--serverpid", dest="serverpid", type='int',
                      help="process id of server to monitor", default=0)
    parser.add_option("-u", "--user", type='string', dest='user',
                      help='name of user to set effective uid to')
    parser.add_option("-i", "--scriptpidfile", type='string', dest='scriptpidfile',
                      help='name of file containing the pid of this script')

    options, args = parser.parse_args()

    args = parse_plugins(parser, options, args)

    if len(args) < 1:
        parser.error("You must specify the name of the pipe to use")
    if len(args) > 1:
        parser.error("error - unhandled command line arguments: %s" % ' '.join(args))

    return options, args[0]

options, logfname = parse_options()

if options.debug:
    debug = True

if len(plgfuncs) == 0:
    plgfuncs.append(defaultplugin)
if len(plgpostfuncs) == 0:
    plgpostfuncs.append(defaultpost)

if options.user:
    try:
        userid = int(options.user)
    except ValueError:  # not a numeric userid - look it up
        try:
            userid = pwd.getpwnam(options.user)[2]
        except Exception as e:
            print("Failed to lookup name (%s) error: %s" %
                  (options.user, str(e)))
            sys.exit(1)
    os.seteuid(userid)

if options.scriptpidfile:
    if not handle_script_pidfile(options.scriptpidfile):
        options.scriptpidfile = None
        sys.exit(1)

serverpid = options.serverpid
if serverpid:
    if not is_proc_alive(serverpid):
        print("Server pid [%d] is not alive - exiting" % serverpid)
        sys.exit(1)

try:
    if os.stat(logfname).st_mode & S_IFIFO:
        if debug:
            print("Using existing log pipe", logfname)
    else:
        print("Error:", logfname, "exists and is not a log pipe")
        print("use a filename other than", logfname)
        sys.exit(1)
except OSError as e:
    if e.errno == errno.ENOENT:
        if debug:
            print("Creating log pipe", logfname)
        try:
            os.mkfifo(logfname)
            os.chmod(logfname, 0o600)
        except Exception as e:
            print("Failed to create log pipe: " + str(e))
            sys.exit(1)
    else:
        print("Failed to create log pipe - %s [error %d]" % (e.strerror, e.errno))
        sys.exit(1)

if debug:
    print("Listening to log pipe", logfname, "number of lines", maxlines)

# set up our signal handlers
signal.signal(signal.SIGHUP, sighandler)
signal.signal(signal.SIGINT, sighandler)
#signal.signal(signal.SIGPIPE, sighandler)
signal.signal(signal.SIGTERM, sighandler)
signal.signal(signal.SIGALRM, sighandler)

timerisset = False
neverdone = False
if options.serverpidfile:
    # start the timer to wait for the pid file to be available
    signal.setitimer(signal.ITIMER_REAL, options.servertimeout)
    timerisset = True

# if we are tracking a server, we will be done
# when the server exits
# if not tracking a server, we will only be done
# when we are killed
if not serverpid and not options.serverpidfile:
    neverdone = True

done = False
while not done:
    # open the pipe - will hang until
    # 1. something opens the other end
    # 2. alarm goes off - will just exit
    logf = open_pipe(logfname)
    # if we get here, logf is not None
    if debug:
        print("opened pipe", logf)

    if timerisset:
        # cancel the timer - the open succeeded
        timerisset = False
        signal.setitimer(signal.ITIMER_REAL, 0)
        if debug:
            print("cancelled startup timer")

    lines = 0
    # read and process the next line in the pipe
    # if server exits while we are reading, we will get
    # EOF and the func will return True - will also
    # return True if a plugin returns failure
    while not read_and_process_line(logf, plgfuncs):
        lines += 1

    # the other end of the pipe closed - we close our end too
    if debug:
        print("read", lines, "lines")
    logf.close()
    logf = None
    if debug:
        print("closed log pipe", logfname)

    if not serverpid and options.serverpidfile:
        # see if the server has written its server pid file yet
        # it may take a "long time" for the server to actually
        # write its pid file
        serverpid = get_pid_from_file(options.serverpidfile)

    # if the server is no longer running, just finish
    if serverpid and not is_proc_alive(serverpid):
        done = True
        if debug:
            print("server pid", serverpid, "exited - script exiting")

    if neverdone:
        done = False
    elif not done:
        if not lines:
            # at startup the server will close the log and reopen it
            # when it does this lines will be 0 - this means we need
            # immediately attempt to reopen the log pipe and read it
            # however, at shutdown, the server will close the log before
            # the process has exited - so is_proc_alive will return
            # true for a short time - if we then attempt to open the
            # pipe, the open will hang forever - to avoid this situation
            # we set the alarm again to wake up the open - use a short
            # timeout so we don't wait a long time if the server
            # really is exiting
            signal.setitimer(signal.ITIMER_REAL, 0.25)
            timerisset = True
            if debug:
                print("set startup timer - see if server is really shut down")
        else: # we read something
            # pipe closed - usually when server shuts down
            done = True

    if not done and debug:
        print("log pipe", logfname, "closed - reopening - read", totallines, "total lines")

finish()



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2021 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import json
import ldap
from lib389.plugins import DNAPlugin, DNAPluginConfig, DNAPluginConfigs, DNAPluginSharedConfig, DNAPluginSharedConfigs
from lib389.cli_conf import add_generic_plugin_parsers, generic_object_edit, generic_object_add, _args_to_attrs
from lib389.cli_base import CustomHelpFormatter

arg_to_attr = {
    'type': 'dnaType',
    'prefix': 'dnaPrefix',
    'next_value': 'dnaNextValue',
    'max_value': 'dnaMaxValue',
    'interval': 'dnaInterval',
    'magic_regen': 'dnaMagicRegen',
    'filter': 'dnaFilter',
    'scope': 'dnaScope',
    'remote_bind_dn': 'dnaRemoteBindDN',
    'remote_bind_cred': 'dnaRemoteBindCred',
    'shared_config_entry': 'dnaSharedCfgDN',
    'threshold': 'dnaThreshold',
    'next_range': 'dnaNextRange',
    'range_request_timeout': 'dnaRangeRequestTimeout'
}

arg_to_attr_config = {
    'HOSTNAME': 'dnaHostname',
    'PORT': 'dnaPortNum',
    'secure_port': 'dnaSecurePortNum',
    'remaining_values': 'dnaRemainingValues',
    'remote_bind_method': 'dnaRemoteBindMethod',
    'remote_conn_protocol': 'dnaRemoteConnProtocol'
}


def _get_shared_config_dn(inst, args):
    configs = DNAPluginConfigs(inst)
    config = configs.get(args.NAME)
    if config.present('dnaSharedCfgDN'):
        basedn = config.get_attr_val_utf8_l('dnaSharedCfgDN')
    else:
        raise ValueError('dnaSharedCfgDN should be set at the "%s" config entry' % args.NAME)
    if ':' not in args.SHARED_CFG:
       raise ValueError("The shared config entry identifier must be HOSTNAME:PORT")
    (hostname, port) = args.SHARED_CFG.split(":", 1)
    decomposed_dn = [[('dnaHostname', hostname, 1),
                      ('dnaPortNum', port, 1)]] + ldap.dn.str2dn(basedn)
    return ldap.dn.dn2str(decomposed_dn)


def dna_list(inst, basedn, log, args):
    log = log.getChild('dna_list')
    configs = DNAPluginConfigs(inst)
    result = []
    result_json = []
    for config in configs.list():
        if args.json:
            result_json.append(json.loads(config.get_all_attrs_json()))
        else:
            result.append(config.rdn)
    if args.json:
        log.info(json.dumps({"type": "list", "items": result_json}, indent=4))
    else:
        if len(result) > 0:
            for i in result:
                log.info(i)
        else:
            log.info("No DNA configurations were found")


def dna_add(inst, basedn, log, args):
    log = log.getChild('dna_add')
    plugin = DNAPlugin(inst)
    props = {'cn': args.NAME}
    generic_object_add(DNAPluginConfig, inst, log, args, arg_to_attr, basedn=plugin.dn, props=props)


def dna_edit(inst, basedn, log, args):
    log = log.getChild('dna_edit')
    configs = DNAPluginConfigs(inst)
    config = configs.get(args.NAME)
    generic_object_edit(config, log, args, arg_to_attr)


def dna_show(inst, basedn, log, args):
    log = log.getChild('dna_show')
    configs = DNAPluginConfigs(inst)
    config = configs.get(args.NAME)

    if not config.exists():
        raise ldap.NO_SUCH_OBJECT("Entry %s doesn't exists" % args.NAME)
    if args and args.json:
        o_str = config.get_all_attrs_json()
        log.info(o_str)
    else:
        log.info(config.display())


def dna_del(inst, basedn, log, args):
    log = log.getChild('dna_del')
    configs = DNAPluginConfigs(inst)
    config = configs.get(args.NAME)
    config.delete()
    log.info("Successfully deleted the %s", config.dn)


def dna_config_list(inst, basedn, log, args):
    log = log.getChild('dna_list')
    configs = DNAPluginSharedConfigs(inst, args.BASEDN)
    result = []
    result_json = []
    parent_config_entries = []

    parent_configs = DNAPluginConfigs(inst)
    for config in parent_configs.list():
        if config.present("dnaSharedCfgDN") and config.get_attr_val_utf8("dnaSharedCfgDN") == args.BASEDN:
            parent_config_entries.append(config.rdn)

    for config in configs.list():
        if args.json:
            result_json.append(json.loads(config.get_all_attrs_json()))
        else:
            port = config.get_attr_val_utf8('dnaPortNum')
            shared_config = f"{config.rdn}:{port}"
            result.append(shared_config)
    if args.json:
        log.info(json.dumps({"type": "list", "items": result_json}, indent=4))
    else:
        if len(result) > 0:
            for i in result:
                log.info(i)
            if parent_config_entries:
                log.info("DNA plugin configs which have the shared config entry set as a dnaSharedCfgDN attribute: " + " ".join(parent_config_entries))
            else:
                log.info("No DNA plugin configs have the shared config entry set as a dnaSharedCfgDN attribute.")
        else:
            log.info("No DNA shared configurations were found")


def dna_config_edit(inst, basedn, log, args):
    log = log.getChild('dna_config_edit')
    targetdn = _get_shared_config_dn(inst, args)

    shared_config = DNAPluginSharedConfig(inst, targetdn)
    generic_object_edit(shared_config, log, args, arg_to_attr_config)


def dna_config_show(inst, basedn, log, args):
    log = log.getChild('dna_config_show')
    targetdn = _get_shared_config_dn(inst, args)

    shared_config = DNAPluginSharedConfig(inst, targetdn)

    if not shared_config.exists():
        raise ldap.NO_SUCH_OBJECT("Entry %s doesn't exists" % targetdn)
    if args and args.json:
        o_str = shared_config.get_all_attrs_json()
        log.info(o_str)
    else:
        log.info(shared_config.display())


def dna_config_del(inst, basedn, log, args):
    log = log.getChild('dna_config_del')
    targetdn = _get_shared_config_dn(inst, args)
    shared_config = DNAPluginSharedConfig(inst, targetdn)
    shared_config.delete()
    log.info("Successfully deleted the %s", targetdn)


def _add_parser_args(parser):
    parser.add_argument('--type', nargs='+',
                        help='Sets which attributes have unique numbers being generated for them (dnaType)')
    parser.add_argument('--prefix', help='Defines a prefix that can be prepended to the generated '
                                         'number values for the attribute (dnaPrefix)')
    parser.add_argument('--next-value', help='Sets the next available number which can be assigned (dnaNextValue)')
    parser.add_argument('--max-value', help='Sets the maximum value that can be assigned for the range (dnaMaxValue)')
    parser.add_argument('--interval', help='Sets an interval to use to increment through numbers in a range (dnaInterval)')
    parser.add_argument('--magic-regen', help='Sets a user-defined value that instructs the plug-in '
                                              'to assign a new value for the entry (dnaMagicRegen)')
    parser.add_argument('--filter', help='Sets an LDAP filter to use to search for and identify the entries '
                                         'to which to apply the distributed numeric assignment range (dnaFilter)')
    parser.add_argument('--scope', help='Sets the base DN to search for entries to which '
                                        'to apply the distributed numeric assignment (dnaScope)')
    parser.add_argument('--remote-bind-dn', help='Specifies the Replication Manager DN (dnaRemoteBindDN)')
    parser.add_argument('--remote-bind-cred', help='Specifies the Replication Manager\'s password (dnaRemoteBindCred)')
    parser.add_argument('--shared-config-entry', help='Defines a shared identity that the servers can use '
                                                      'to transfer ranges to one another (dnaSharedCfgDN)')
    parser.add_argument('--threshold', help='Sets a threshold of remaining available numbers in the range. When the '
                                            'server hits the threshold, it sends a request for a new range (dnaThreshold)')
    parser.add_argument('--next-range',
                        help='Defines the next range to use when the current range is exhausted (dnaNextRange)')
    parser.add_argument('--range-request-timeout',
                        help='Sets a timeout period, in seconds, for range requests so that the server '
                             'does not stall waiting on a new range from one server and '
                             'can request a range from a new server (dnaRangeRequestTimeout)')

def create_parser(subparsers):
    dna = subparsers.add_parser('dna', help='Manage and configure DNA plugin', formatter_class=CustomHelpFormatter)
    subcommands = dna.add_subparsers(help='action')
    add_generic_plugin_parsers(subcommands, DNAPlugin)

    list = subcommands.add_parser('list', help='List available plugin configs', formatter_class=CustomHelpFormatter)
    subcommands_list = list.add_subparsers(help='action')
    list_configs = subcommands_list.add_parser('configs', help='List main DNA plugin config entries', formatter_class=CustomHelpFormatter)
    list_configs.set_defaults(func=dna_list)
    list_shared_configs = subcommands_list.add_parser('shared-configs', help='List DNA plugin shared config entries', formatter_class=CustomHelpFormatter)
    list_shared_configs.add_argument('BASEDN', help='The search DN')
    list_shared_configs.set_defaults(func=dna_config_list)

    config = subcommands.add_parser('config', help='Manage plugin configs', formatter_class=CustomHelpFormatter)
    config.add_argument('NAME', help='The DNA configuration name')
    config_subcommands = config.add_subparsers(help='action')
    add = config_subcommands.add_parser('add', help='Add the config entry', formatter_class=CustomHelpFormatter)
    add.set_defaults(func=dna_add)
    _add_parser_args(add)
    edit = config_subcommands.add_parser('set', help='Edit the config entry', formatter_class=CustomHelpFormatter)
    edit.set_defaults(func=dna_edit)
    _add_parser_args(edit)
    show = config_subcommands.add_parser('show', help='Display the config entry', formatter_class=CustomHelpFormatter)
    show.set_defaults(func=dna_show)
    delete = config_subcommands.add_parser('delete', help='Delete the config entry', formatter_class=CustomHelpFormatter)
    delete.set_defaults(func=dna_del)

    shared_config = config_subcommands.add_parser('shared-config-entry', help='Manage the shared config entry', formatter_class=CustomHelpFormatter)
    shared_config.add_argument('SHARED_CFG',
                               help='Use HOSTNAME:PORT for this argument to identify the host name and port of a server in a shared range, as part of the DNA range '
                                    'configuration for that specific host in multi-supplier replication.  (dnaHostname+dnaPortNum)')
    shared_config_subcommands = shared_config.add_subparsers(help='action')
    edit_config = shared_config_subcommands.add_parser('set', help='Edit the shared config entry', formatter_class=CustomHelpFormatter)
    edit_config.set_defaults(func=dna_config_edit)
    edit_config.add_argument('--remote-bind-method', help='Specifies the remote bind method "SIMPLE", "SSL" (for SSL client auth), "SASL/GSSAPI", or "SASL/DIGEST-MD5" (dnaRemoteBindMethod)')
    edit_config.add_argument('--remote-conn-protocol', help='Specifies the remote connection protocol "LDAP", or "TLS" (dnaRemoteConnProtocol)')

    show_config_parser = shared_config_subcommands.add_parser('show', help='Display the shared config entry', formatter_class=CustomHelpFormatter)
    show_config_parser.set_defaults(func=dna_config_show)
    del_config_parser = shared_config_subcommands.add_parser('delete', help='Delete the shared config entry', formatter_class=CustomHelpFormatter)
    del_config_parser.set_defaults(func=dna_config_del)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2023 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import json
from os.path import expanduser
from os import path, remove
from ldapurl import isLDAPUrl
from ldap.dn import is_dn
import configparser
from lib389.utils import is_a_dn
from lib389.cli_base import CustomHelpFormatter


def create_dsrc(inst, log, args):
    """Create the .dsrc file

    [instance]
    uri = ldaps://hostname:port
    basedn = dc=example,dc=com
    people_rdn = ou=people
    groups_rdn = ou=groups
    binddn = uid=user,....
    saslmech = [EXTERNAL|PLAIN]
    tls_cacertdir = /path/to/cacertdir
    tls_cert = /path/to/user.crt
    tls_key = /path/to/user.key
    tls_reqcert = [never, hard, allow]
    starttls = [true, false]
    pwdfile = /path/to/file
    """

    dsrc_file = f'{expanduser("~")}/.dsrc'
    config = configparser.ConfigParser()
    config.read(dsrc_file)

    # Verify this section does not already exist
    instances = config.sections()
    if inst.serverid in instances:
        raise ValueError("There is already a configuration section for this instance!")

    # Process and validate the args
    config[inst.serverid] = {}

    if args.uri is not None:
        if not isLDAPUrl(args.uri):
            raise ValueError("The uri is not a valid LDAP URL!")
        if args.uri.startswith("ldapi"):
            # We must use EXTERNAL saslmech for LDAPI
            args.saslmech = "EXTERNAL"
        config[inst.serverid]['uri'] = args.uri
    if args.basedn is not None:
        if not is_dn(args.basedn):
            raise ValueError("The basedn is not a valid DN!")
        config[inst.serverid]['basedn'] = args.basedn
    if args.binddn is not None:
        if not is_dn(args.binddn):
            raise ValueError("The binddn is not a valid DN!")
        config[inst.serverid]['binddn'] = args.binddn
    if args.saslmech is not None:
        if args.saslmech not in ['EXTERNAL', 'PLAIN']:
            raise ValueError("The saslmech must be EXTERNAL or PLAIN!")
        config[inst.serverid]['saslmech'] = args.saslmech
    if args.tls_cacertdir is not None:
        if not path.exists(args.tls_cacertdir):
            raise ValueError('--tls-cacertdir directory does not exist!')
        config[inst.serverid]['tls_cacertdir'] = args.tls_cacertdir
    if args.tls_cert is not None:
        if not path.exists(args.tls_cert):
            raise ValueError('--tls-cert does not point to an existing file!')
        config[inst.serverid]['tls_cert'] = args.tls_cert
    if args.tls_key is not None:
        if not path.exists(args.tls_key):
            raise ValueError('--tls-key does not point to an existing file!')
        config[inst.serverid]['tls_key'] = args.tls_key
    if args.tls_reqcert is not None:
        if args.tls_reqcert not in ['never', 'hard', 'allow']:
            raise ValueError('--tls-reqcert value is invalid (must be either "never", "allow", or "hard")!')
        config[inst.serverid]['tls_reqcert'] = args.tls_reqcert
    if args.starttls:
         config[inst.serverid]['starttls'] = 'true'
    if args.pwdfile is not None:
        if not path.exists(args.pwdfile):
            raise ValueError('--pwdfile does not exist!')
        config[inst.serverid]['pwdfile'] = args.pwdfile
    if args.people_rdn is not None:
        if not is_dn(args.people_rdn):
            raise ValueError("The value for people-dn is not a valid DN!")
        config[inst.serverid]['people_rdn'] = args.people_rdn
    if args.groups_rdn is not None:
        if not is_dn(args.groups_rdn):
            raise ValueError("The value for groups-dn is not a valid DN!")
        config[inst.serverid]['groups_rdn'] = args.groups_rdn

    if len(config[inst.serverid]) == 0:
        # No args set
        raise ValueError("You must set at least one argument for the new dsrc file!")

    # Print a preview of the config
    log.info(f'Updating "{dsrc_file}" with:\n')
    log.info(f'    [{inst.serverid}]')
    for k, v in config[inst.serverid].items():
        log.info(f'    {k} = {v}')

    # Perform confirmation?
    if not args.do_it:
        while 1:
            val = input(f'\nUpdate "{dsrc_file}" ? [yes]: ').rstrip().lower()
            if val == '' or val == 'y' or val == 'yes':
                break
            if val == 'n' or val == 'no':
                return

    # Now write the file
    with open(dsrc_file, 'w') as configfile:
        config.write(configfile)

    log.info(f'Successfully updated: {dsrc_file}')


def modify_dsrc(inst, log, args):
    """Modify the instance config
    """
    dsrc_file = f'{expanduser("~")}/.dsrc'

    if path.exists(dsrc_file):
        config = configparser.ConfigParser()
        config.read(dsrc_file)

        # Verify we have a section to modify
        instances = config.sections()
        if inst.serverid not in instances:
            raise ValueError("There is no configuration section for this instance to modify!")

        # Process and validate the args
        if args.uri is not None:
            if not isLDAPUrl(args.uri):
                raise ValueError("The uri is not a valid LDAP URL!")
            if args.uri.startswith("ldapi"):
                # We must use EXTERNAL saslmech for LDAPI
                args.saslmech = "EXTERNAL"
            if args.uri == '':
                del config[inst.serverid]['uri']
            else:
                config[inst.serverid]['uri'] = args.uri
        if args.basedn is not None:
            if not is_dn(args.basedn):
                raise ValueError("The basedn is not a valid DN!")
            if args.basedn == '':
                del config[inst.serverid]['basedn']
            else:
                config[inst.serverid]['basedn'] = args.basedn
        if args.binddn is not None:
            if not is_dn(args.binddn):
                raise ValueError("The binddn is not a valid DN!")
            if args.binddn == '':
                del config[inst.serverid]['binddn']
            else:
                config[inst.serverid]['binddn'] = args.binddn
        if args.saslmech is not None:
            if args.saslmech not in ['EXTERNAL', 'PLAIN']:
                raise ValueError("The saslmech must be EXTERNAL or PLAIN!")
            if args.saslmech == '':
                del config[inst.serverid]['saslmech']
            else:
                config[inst.serverid]['saslmech'] = args.saslmech
        if args.tls_cacertdir is not None:
            if not path.exists(args.tls_cacertdir):
                raise ValueError('--tls-cacertdir directory does not exist!')
            if args.tls_cacertdir == '':
                del config[inst.serverid]['tls_cacertdir']
            else:
                config[inst.serverid]['tls_cacertdir'] = args.tls_cacertdir
        if args.tls_cert is not None:
            if not path.exists(args.tls_cert):
                raise ValueError('--tls-cert does not point to an existing file!')
            if args.tls_cert == '':
                del config[inst.serverid]['tls_cert']
            else:
                config[inst.serverid]['tls_cert'] = args.tls_cert
        if args.tls_key is not None:
            if not path.exists(args.tls_key):
                raise ValueError('--tls-key does not point to an existing file!')
            if args.tls_key == '':
                del config[inst.serverid]['tls_key']
            else:
                config[inst.serverid]['tls_key'] = args.tls_key
        if args.tls_reqcert is not None:
            if args.tls_reqcert not in ['never', 'hard', 'allow']:
                raise ValueError('--tls-reqcert value is invalid (must be either "never", "allow", or "hard")!')
            if args.tls_reqcert == '':
                del config[inst.serverid]['tls_reqcert']
            else:
                config[inst.serverid]['tls_reqcert'] = args.tls_reqcert
        if args.starttls:
             config[inst.serverid]['starttls'] = 'true'
        if args.cancel_starttls:
            # coverity[copy_paste_error] does not work as expected
            # so lets work around the issue by splitting the string
            config[inst.serverid]['star' + 'ttls'] = 'false'
        if args.pwdfile is not None:
            if not path.exists(args.pwdfile):
                raise ValueError('--pwdfile does not exist!')
            if args.pwdfile == '':
                del config[inst.serverid]['pwdfile']
            else:
                config[inst.serverid]['pwdfile'] = args.pwdfile
        if args.people_rdn is not None:
            if not is_dn(args.people_rdn):
                raise ValueError('--people-rdn is not a valid DN!')
            if args.people_rdn == '':
                del config[inst.serverid]['people_rdn']
            else:
                config[inst.serverid]['people_rdn'] = args.people_rdn
        if args.groups_rdn is not None:
            if not is_dn(args.groups_rdn):
                raise ValueError('--groups-rdn is not a valid DN!')
            if args.groups_rdn == '':
                del config[inst.serverid]['groups_rdn']
            else:
                config[inst.serverid]['groups_rdn'] = args.groups_rdn

        # Okay now rewrite the file
        with open(dsrc_file, 'w') as configfile:
            config.write(configfile)

        log.info(f'Successfully updated: {dsrc_file}')
    else:
        raise ValueError(f'There is no .dsrc file "{dsrc_file}" to update!')


def delete_dsrc(inst, log, args):
    """Delete the .dsrc file
    """
    dsrc_file = f'{expanduser("~")}/.dsrc'
    if path.exists(dsrc_file):
        if not args.do_it:
            # Get confirmation
            while 1:
                val = input(f'\nAre you sure you want to remove this instances configuration ? [no]: ').rstrip().lower()
                if val == 'y' or val == 'yes':
                    break
                if val == '' or val == 'n' or val == 'no':
                    return

        config = configparser.ConfigParser()
        config.read(dsrc_file)
        instances = config.sections()
        if inst.serverid not in instances:
            raise ValueError("The is no configuration for this instance")

        # Update the config object
        del config[inst.serverid]

        if len(config.sections()) == 0:
            # The file would be empty so just delete it
            try:
                remove(dsrc_file)
                log.info(f'Successfully removed: {dsrc_file}')
                return
            except OSError as e:
                raise ValueError(f'Failed to delete "{dsrc_file}",  error: {str(e)}')
        else:
            # write the updated config
            with open(dsrc_file, 'w') as configfile:
                config.write(configfile)
    else:
        raise ValueError(f'There is no .dsrc file "{dsrc_file}" to update!')

    log.info(f'Successfully updated: {dsrc_file}')


def display_dsrc(inst, log, args):
    """Display the contents of the ~/.dsrc file
    """
    dsrc_file = f'{expanduser("~")}/.dsrc'

    if not path.exists(dsrc_file):
        raise ValueError(f'There is no dsrc file "{dsrc_file}" to display!')

    config = configparser.ConfigParser()
    config.read(dsrc_file)
    instances = config.sections()

    result = {}
    for inst_section in instances:
        if args.json:
            result[inst_section] = dict(config[inst_section])
        else:
            log.info(f'[{inst_section}]')
            for k, v in config[inst_section].items():
                log.info(f'{k} = {v}')
            log.info("")

    if args.json:
        log.info(json.dumps(result, indent=4))


def replmon_dsrc(inst, log, args):
    dsrc_file = f'{expanduser("~")}/.dsrc'
    repl_conn_section = 'repl-monitor-connections'
    repl_alias_section = 'repl-monitor-aliases'

    config = configparser.ConfigParser()
    config.read(dsrc_file)

    # Process and validate the args
    if args.add_conn is not None:
        sections = config.sections()
        if repl_conn_section not in sections:
            config.add_section(repl_conn_section)
        for conn in args.add_conn:
            conn_parts = conn.split(":")
            if len(conn_parts) != 5:
                raise ValueError("Missing required information for connection: NAME:HOST:PORT:BINDDN:CREDENTIAL")
            conn_name = conn_parts[0]
            conn_host = conn_parts[1]
            conn_port = conn_parts[2]
            conn_binddn = conn_parts[3]
            conn_cred = conn_parts[4]
            if type(int(conn_port)) != int:
                raise ValueError("Invalid value for PORT, must be a number")
            if not is_a_dn(conn_binddn):
                raise ValueError("The bind DN is not a valid DN")
            if conn_name in config[repl_conn_section]:
                raise ValueError("Replication connection with the same name already exists")
            config[repl_conn_section][conn_name] = f"{conn_host}:{conn_port}:{conn_binddn}:{conn_cred}"

    if args.del_conn is not None:
        sections = config.sections()
        if repl_conn_section not in sections:
            raise ValueError("There are no connections configured")
        for conn in args.del_conn:
            if conn not in config.options(repl_conn_section):
                raise ValueError("Can not find connection: " + conn)
            del config[repl_conn_section][conn]
            if len(config[repl_conn_section]) == 0:
                # Delete section header
                del config[repl_conn_section]

    if args.add_alias is not None:
        sections = config.sections()
        if repl_alias_section not in sections:
            config.add_section(repl_alias_section)
        for alias in args.add_alias:
            alias_parts = alias.split(":")
            if len(alias_parts) != 3:
                raise ValueError("Missing required information for alias: ALIAS_NAME:HOST:PORT")
            alias_name = alias_parts[0]
            alias_host = alias_parts[1]
            alias_port = alias_parts[2]
            if type(int(alias_port)) != int:
                raise ValueError("Invalid value for PORT, must be a number")
            if alias_name in config[repl_alias_section]:
                raise ValueError("Replica alias with the same name already exists")
            config[repl_alias_section][alias_name] = f"{alias_host}:{alias_port}"

    if args.del_alias is not None:
        sections = config.sections()
        if repl_alias_section not in sections:
            raise ValueError("There are no aliases configured")
        for alias in args.del_alias:
            if alias not in config.options(repl_alias_section):
                raise ValueError("Can not find alias: " + alias)
            del config[repl_alias_section][alias]
            if len(config[repl_alias_section]) == 0:
                # delete section header
                del config[repl_alias_section]

    # Okay now rewrite the file
    with open(dsrc_file, 'w') as configfile:
        config.write(configfile)

    log.info(f'Successfully updated: {dsrc_file}')


def create_parser(subparsers):
    dsrc_parser = subparsers.add_parser('dsrc', help="Manage the .dsrc file", formatter_class=CustomHelpFormatter)
    subcommands = dsrc_parser.add_subparsers(help="action")

    # Create .dsrc file
    dsrc_create_parser = subcommands.add_parser('create', help='Generate the .dsrc file', formatter_class=CustomHelpFormatter)
    dsrc_create_parser.set_defaults(func=create_dsrc)
    dsrc_create_parser.add_argument('--uri', help="The URI (LDAP URL) for the Directory Server instance.")
    dsrc_create_parser.add_argument('--basedn', help="The default database suffix.")
    dsrc_create_parser.add_argument('--people-rdn',
                                    help="Set the RDN for the 'people' subtree.  Default is \"ou=people\"")
    dsrc_create_parser.add_argument('--groups-rdn',
                                    help="Set the RDN for the 'groups' subtree.  Default is \"ou=groups\"")
    dsrc_create_parser.add_argument('--binddn', help="The default Bind DN used or authentication.")
    dsrc_create_parser.add_argument('--saslmech', help="The SASL mechanism to use: PLAIN or EXTERNAL.")
    dsrc_create_parser.add_argument('--tls-cacertdir', help="The directory containing the Trusted Certificate Authority certificate.")
    dsrc_create_parser.add_argument('--tls-cert', help="The absolute file name to the server certificate.")
    dsrc_create_parser.add_argument('--tls-key', help="The absolute file name to the server certificate key.")
    dsrc_create_parser.add_argument('--tls-reqcert', help="Request certificate strength: 'never', 'allow', 'hard'")
    dsrc_create_parser.add_argument('--starttls', action='store_true', help="Use startTLS for connection to the server.")
    dsrc_create_parser.add_argument('--pwdfile', help="The absolute path to a file containing the Bind DN's password.")
    dsrc_create_parser.add_argument('--do-it', action='store_true', help="Create the file without any confirmation.")

    dsrc_modify_parser = subcommands.add_parser('modify', help='Modify the .dsrc file', formatter_class=CustomHelpFormatter)
    dsrc_modify_parser.set_defaults(func=modify_dsrc)
    dsrc_modify_parser.add_argument('--uri', nargs='?', const='', help="The URI (LDAP URL) for the Directory Server instance.")
    dsrc_modify_parser.add_argument('--basedn', nargs='?', const='', help="The default database suffix.")
    dsrc_modify_parser.add_argument('--people-rdn', nargs='?', const='', help="Sets the RDN used for the 'people' container")
    dsrc_modify_parser.add_argument('--groups-rdn', nargs='?', const='', help="Sets the RDN used for the 'groups' container")
    dsrc_modify_parser.add_argument('--binddn', nargs='?', const='', help="The default Bind DN used or authentication.")
    dsrc_modify_parser.add_argument('--saslmech', nargs='?', const='', help="The SASL mechanism to use: PLAIN or EXTERNAL.")
    dsrc_modify_parser.add_argument('--tls-cacertdir', nargs='?', const='', help="The directory containing the Trusted Certificate Authority certificate.")
    dsrc_modify_parser.add_argument('--tls-cert', nargs='?', const='', help="The absolute file name to the server certificate.")
    dsrc_modify_parser.add_argument('--tls-key', nargs='?', const='', help="The absolute file name to the server certificate key.")
    dsrc_modify_parser.add_argument('--tls-reqcert', nargs='?', const='', help="Request certificate strength: 'never', 'allow', 'hard'")
    dsrc_modify_parser.add_argument('--starttls', action='store_true', help="Use startTLS for connection to the server.")
    dsrc_modify_parser.add_argument('--cancel-starttls', action='store_true', help="Do not use startTLS for connection to the server.")
    dsrc_modify_parser.add_argument('--pwdfile', nargs='?', const='', help="The absolute path to a file containing the Bind DN's password.")
    dsrc_modify_parser.add_argument('--do-it', action='store_true', help="Update the file without any confirmation.")

    # Delete the instance from the .dsrc file
    dsrc_delete_parser = subcommands.add_parser('delete', help='Delete instance configuration from the .dsrc file.', formatter_class=CustomHelpFormatter)
    dsrc_delete_parser.set_defaults(func=delete_dsrc)
    dsrc_delete_parser.add_argument('--do-it', action='store_true',
                                    help="Delete this instance's configuration from the .dsrc file.")

    # Display .dsrc file
    dsrc_display_parser = subcommands.add_parser('display', help='Display the contents of the .dsrc file.', formatter_class=CustomHelpFormatter)
    dsrc_display_parser.set_defaults(func=display_dsrc)

    # Replication Monitor actions
    dsrc_replmon_parser = subcommands.add_parser('repl-mon', help='Display the contents of the .dsrc file.', formatter_class=CustomHelpFormatter)
    dsrc_replmon_parser.set_defaults(func=replmon_dsrc)
    dsrc_replmon_parser.add_argument('--add-conn', nargs='+', help="Add a replica connection: 'NAME:HOST:PORT:BINDDN:CREDENTIAL'")
    dsrc_replmon_parser.add_argument('--del-conn', nargs='+', help="delete a replica connection by its NAME")
    dsrc_replmon_parser.add_argument('--add-alias', nargs='+', help="Add a host/port alias: 'ALIAS_NAME:HOST:PORT'")
    dsrc_replmon_parser.add_argument('--del-alias', nargs='+', help="delete a host/port alias by its ALIAS_NAME")



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2016, William Brown <william at blackhats.net.au>
# Copyright (C) 2024 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

from lib389._mapped_object import DSLdapObject, DSLdapObjects
from lib389.utils import ds_is_older, ensure_str
from lib389.cli_base.dsrc import dsrc_to_ldap
from lib389._constants import DSRC_HOME

MUST_ATTRIBUTES = [
    'cn',
    'gidNumber',
]
RDN = 'cn'
DEFAULT_BASEDN_RDN = 'ou=Groups'


class PosixGroup(DSLdapObject):
    """A single instance of PosixGroup entry

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """

    def __init__(self, instance, dn=None):
        super(PosixGroup, self).__init__(instance, dn)
        self._rdn_attribute = RDN
        # Can I generate these from schema?
        self._must_attributes = MUST_ATTRIBUTES
        self._create_objectclasses = [
            'top',
            'groupOfNames',
            'posixGroup',
        ]
        if not ds_is_older('1.3.7', instance=instance):
            self._create_objectclasses.append('nsMemberOf')
        self._protected = False

    def check_member(self, dn):
        """Check if DN is a member

        :param dn: Entry DN
        :type dn: str
        """

        return dn in self.get_attr_vals('member')

    def add_member(self, dn):
        """Add DN as a member

        :param dn: Entry DN
        :type dn: str
        """

        # Assert the DN exists?
        self.add('member', dn)


class PosixGroups(DSLdapObjects):
    """DSLdapObjects that represents PosixGroups entry
    By default it uses 'ou=Groups' as rdn.

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param basedn: Base DN for all group entries below
    :type basedn: str
    """

    def __init__(self, instance, basedn, rdn=DEFAULT_BASEDN_RDN):
        super(PosixGroups, self).__init__(instance)
        self._objectclasses = [
            'groupOfNames',
            'posixGroup',
        ]
        self._filterattrs = [RDN]
        self._childobject = PosixGroup

        dsrc_inst = dsrc_to_ldap(DSRC_HOME, instance.serverid, self._log)
        if dsrc_inst is not None and 'groups_rdn' in dsrc_inst and dsrc_inst['groups_rdn'] is not None:
            rdn = dsrc_inst['groups_rdn']

        if rdn is None:
            self._basedn = ensure_str(basedn)
        else:
            self._basedn = '{},{}'.format(ensure_str(rdn), ensure_str(basedn))



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2016, William Brown <william at blackhats.net.au>
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

from lib389._mapped_object import DSLdapObject

class Domain(DSLdapObject):
    """A single instance of Domain entry
    - must attributes = ['dc']
    - RDN attribute is 'dc'

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Entry DN
    :type dn: str
    """

    def __init__(self, instance, dn):
        super(Domain, self).__init__(instance, dn)
        self._rdn_attribute = 'dc'
        self._must_attributes = ['dc']
        self._create_objectclasses = [
            'top',
            'domain',
        ]
        self._protected = True




# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2025 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import copy
import ldap
import psutil
from lib389._constants import *
from lib389._mapped_object import DSLdapObject
from lib389.utils import (ds_is_older, get_mount_point, get_disk_space)
from lib389.lint import DSDSLE0001
from lib389.backend import DatabaseConfig


class Monitor(DSLdapObject):
    """An object that helps reading of cn=monitor for server statistics.
        :param instance: An instance
        :type instance: lib389.DirSrv
        :param dn: not used
    """
    def __init__(self, instance, dn=None):
        super(Monitor, self).__init__(instance=instance)
        self._dn = DN_MONITOR

    def get_connections(self):
        """Get connection related attribute values for cn=monitor

        :returns: Values of connection, currentconnections,
                  totalconnections attributes of cn=monitor
        """
        connection = self.get_attr_vals_utf8('connection')
        currentconnections = self.get_attr_vals_utf8('currentconnections')
        totalconnections = self.get_attr_vals_utf8('totalconnections')
        return (connection, currentconnections, totalconnections)

    def get_version(self):
        """Get version attribute value for cn=monitor

        :returns: Value of version attribute of cn=monitor
        """
        version = self.get_attr_vals_utf8('connection')
        return version

    def get_threads(self):
        """Get thread related attributes value for cn=monitor

        :returns: Values of threads, currentconnectionsatmaxthreads, and
                  maxthreadsperconnhits attributes of cn=monitor
        """
        threads = self.get_attr_vals_utf8('threads')
        currentconnectionsatmaxthreads = self.get_attr_vals_utf8('currentconnectionsatmaxthreads')
        maxthreadsperconnhits = self.get_attr_vals_utf8('maxthreadsperconnhits')
        return (threads, currentconnectionsatmaxthreads, maxthreadsperconnhits)

    def get_work_queue(self):
        """Get work queue related attributes value for cn=monitor

        :returns: Values of currentworkqueue, maxworkqueue,
                  currentbusyworkers, maxbusyworkers attributes of cn=monitor
        """
        currentworkqueue = self.get_attr_vals_utf8('currentworkqueue')
        maxworkqueue = self.get_attr_vals_utf8('maxworkqueue')
        currentbusyworkers = self.get_attr_vals_utf8('currentbusyworkers')
        maxbusyworkers = self.get_attr_vals_utf8('maxbusyworkers')
        return (currentworkqueue, maxworkqueue, currentbusyworkers, maxbusyworkers)

    def get_thread_pool_workers(self):
        """Get sanitized per-worker thread pool status values from cn=monitor

        :returns: Values of threadpoolworker attribute of cn=monitor
        """
        return self.get_attr_vals_utf8('threadpoolworker')

    def get_backends(self):
        """Get backends related attributes value for cn=monitor

        :returns: Values of nbackends and backendmonitordn attributes of cn=monitor
        """
        nbackends = self.get_attr_vals_utf8('nbackends')
        backendmonitordn = self.get_attr_vals_utf8('backendmonitordn')
        return (nbackends, backendmonitordn)

    def get_operations(self):
        """Get operations related attributes value for cn=monitor

        :returns: Values of opsinitiated and opscompleted attributes of cn=monitor
        """
        opsinitiated = self.get_attr_vals_utf8('opsinitiated')
        opscompleted = self.get_attr_vals_utf8('opsinitiated')
        return (opsinitiated, opscompleted)

    def get_statistics(self):
        """Get statistics attributes value for cn=monitor

        :returns: Values of dtablesize, readwaiters, entriessent,
                  bytessent, currenttime, starttime attributes of cn=monitor
        """
        dtablesize = self.get_attr_vals_utf8('dtablesize')
        readwaiters = self.get_attr_vals_utf8('readwaiters')
        entriessent = self.get_attr_vals_utf8('entriessent')
        bytessent = self.get_attr_vals_utf8('bytessent')
        currenttime = self.get_attr_vals_utf8('currenttime')
        starttime = self.get_attr_vals_utf8('starttime')
        return (dtablesize, readwaiters, entriessent, bytessent, currenttime, starttime)

    def get_resource_stats(self):
        """
        Get CPU and memory stats
        """
        stats = {}
        try:
            pid = self._instance.get_pid()
        except Exception:
            pid = None
        total_mem = psutil.virtual_memory()[0]

        # Always include total system memory
        stats['total_mem'] = [str(total_mem)]

        # Process-specific stats - only if process is running (pid is not None)
        if pid is not None:
            try:
                p = psutil.Process(pid)
                memory_stats = p.memory_full_info()

                # Get memory & CPU stats
                stats['rss'] = [str(memory_stats[0])]
                stats['vms'] = [str(memory_stats[1])]
                stats['swap'] = [str(memory_stats[9])]
                stats['mem_rss_percent'] = [str(round(p.memory_percent("rss")))]
                stats['mem_vms_percent'] = [str(round(p.memory_percent("vms")))]
                stats['mem_swap_percent'] = [str(round(p.memory_percent("swap")))]
                stats['total_threads'] = [str(p.num_threads())]
                stats['cpu_usage'] = [str(round(p.cpu_percent(interval=0.1)))]
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                # Process exists in PID file but is not accessible or doesn't exist
                pid = None

        # If no valid PID, provide zero values for process stats
        if pid is None:
            stats['rss'] = ['0']
            stats['vms'] = ['0']
            stats['swap'] = ['0']
            stats['mem_rss_percent'] = ['0']
            stats['mem_vms_percent'] = ['0']
            stats['mem_swap_percent'] = ['0']
            stats['total_threads'] = ['0']
            stats['cpu_usage'] = ['0']
            stats['server_status'] = ['PID unavailable']
        else:
            stats['server_status'] = ['Server running']

        # Connections to DS
        if self._instance.port == "0":
            port = "ignore"
        else:
            port = str(self._instance.port)
        if self._instance.sslport == "0":
            sslport = "ignore"
        else:
            sslport = str(self._instance.sslport)

        conn_count = 0
        conn_established_count = 0
        conn_close_wait_count = 0
        conn_time_wait_count = 0
        conns = psutil.net_connections()
        for conn in conns:
            if len(conn[4]) > 0:
                conn_port = str(conn[4][1])
                if conn_port in (port, sslport):
                    if conn[5] == 'TIME_WAIT':
                        conn_time_wait_count += 1
                    if conn[5] == 'CLOSE_WAIT':
                        conn_close_wait_count += 1
                    if conn[5] == 'ESTABLISHED':
                        conn_established_count += 1
                    conn_count += 1

        stats['connection_count'] = [str(conn_count)]
        stats['connection_established_count'] = [str(conn_established_count)]
        stats['connection_close_wait_count'] = [str(conn_close_wait_count)]
        stats['connection_time_wait_count'] = [str(conn_time_wait_count)]

        return stats

    def get_status(self, just_resources=False, use_json=False, ):
        stats = self.get_resource_stats()
        if just_resources:
            return stats

        status = self.get_attrs_vals_utf8([
            'version',
            'threads',
            'connection',
            'currentconnections',
            'totalconnections',
            'currentconnectionsatmaxthreads',
            'maxthreadsperconnhits',
            'dtablesize',
            'readwaiters',
            'opsinitiated',
            'opscompleted',
            'entriessent',
            'bytessent',
            'currenttime',
            'starttime',
            'nbackends',
            'currentworkqueue',
            'maxworkqueue',
            'currentbusyworkers',
            'maxbusyworkers',
            'threadpoolworker',
        ])
        status.update(stats)

        return status


class MonitorLDBM(DSLdapObject):
    """An object that helps reading the global database statistics.
        :param instance: An instance
        :type instance: lib389.DirSrv
        :param dn: not used
    """
    DB_KEYS = {
        DB_IMPL_BDB: [
            'dbcachehits', 'dbcachetries', 'dbcachehitratio',
            'dbcachepagein', 'dbcachepageout', 'dbcacheroevict',
            'dbcacherwevict'
        ],
        DB_IMPL_MDB: [
            'normalizeddncachetries', 'normalizeddncachehits',
            'normalizeddncachemisses', 'normalizeddncachehitratio',
            'normalizeddncacheevictions', 'currentnormalizeddncachesize',
            'maxnormalizeddncachesize', 'currentnormalizeddncachecount',
            'normalizeddncachethreadsize', 'normalizeddncachethreadslots'
        ]
    }
    DB_MONITOR_KEYS = {
        DB_IMPL_BDB: [
            'nsslapd-db-abort-rate', 'nsslapd-db-active-txns', 'nsslapd-db-cache-hit',
            'nsslapd-db-cache-try', 'nsslapd-db-cache-region-wait-rate',
            'nsslapd-db-cache-size-bytes', 'nsslapd-db-clean-pages', 'nsslapd-db-commit-rate',
            'nsslapd-db-deadlock-rate', 'nsslapd-db-dirty-pages', 'nsslapd-db-hash-buckets',
            'nsslapd-db-hash-elements-examine-rate', 'nsslapd-db-hash-search-rate',
            'nsslapd-db-lock-conflicts', 'nsslapd-db-lock-region-wait-rate',
            'nsslapd-db-lock-request-rate', 'nsslapd-db-lockers', 'nsslapd-db-configured-locks',
            'nsslapd-db-current-locks', 'nsslapd-db-max-locks', 'nsslapd-db-current-lock-objects',
            'nsslapd-db-max-lock-objects', 'nsslapd-db-log-bytes-since-checkpoint',
            'nsslapd-db-log-region-wait-rate', 'nsslapd-db-log-write-rate',
            'nsslapd-db-longest-chain-length', 'nsslapd-db-page-create-rate',
            'nsslapd-db-page-read-rate', 'nsslapd-db-page-ro-evict-rate',
            'nsslapd-db-page-rw-evict-rate', 'nsslapd-db-page-trickle-rate',
            'nsslapd-db-page-write-rate', 'nsslapd-db-pages-in-use',
            'nsslapd-db-txn-region-wait-rate', 'nsslapd-db-mp-pagesize'
        ],
        DB_IMPL_MDB: [
            'dbenvmapmaxsize', 'dbenvmapsize', 'dbenvlastpageno',
            'dbenvlasttxnid', 'dbenvmaxreaders', 'dbenvnumreaders',
            'dbenvnumdbis', 'waitingrwtxn', 'activerwtxn',
            'abortrwtxn', 'commitrwtxn', 'granttimerwtxn',
            'lifetimerwtxn', 'waitingrotxn', 'activerotxn',
            'abortrotxn', 'commitrotxn', 'granttimerotxn',
            'lifetimerotxn'
        ]
    }

    def __init__(self, instance, dn=None):
        super(MonitorLDBM, self).__init__(instance=instance)
        self._dn = DN_MONITOR_LDBM
        self._db_mon = MonitorDatabase(instance)
        self.inst_db_impl = self._instance.get_db_lib()
        self._backend_keys = list(self.DB_KEYS.get(self.inst_db_impl, []))
        self._db_mon_keys = list(self.DB_MONITOR_KEYS.get(self.inst_db_impl, []))

        if self.inst_db_impl == DB_IMPL_BDB and not ds_is_older("1.4.0", instance=instance):
            self._backend_keys.extend([
                'normalizeddncachetries', 'normalizeddncachehits',
                'normalizeddncachemisses', 'normalizeddncachehitratio',
                'normalizeddncacheevictions', 'currentnormalizeddncachesize',
                'maxnormalizeddncachesize', 'currentnormalizeddncachecount',
                'normalizeddncachethreadsize', 'normalizeddncachethreadslots'
            ])

    def get_status(self, use_json=False):
        ldbm_dict = self.get_attrs_vals_utf8(self._backend_keys)
        db_dict = self._db_mon.get_attrs_vals_utf8(self._db_mon_keys)
        return {**ldbm_dict, **db_dict}


class MonitorDatabase(DSLdapObject):
    """An object that helps reading the global libdb(bdb) or libdb(mdb) statistics.
        :param instance: An instance
        :type instance: lib389.DirSrv
        :param dn: not used
    """
    def __init__(self, instance, dn=None):
        super(MonitorDatabase, self).__init__(instance=instance)
        self._dn = DN_MONITOR_DATABASE
        self._backend_keys = None

    def __init2(self):
        # Determine the key when really accessing the object with get_status
        # because config attrbute and connectio are not yet set in DirSrv
        # when __init is called
        db_lib = self._instance.get_db_lib()
        if db_lib == "bdb":
            self._backend_keys = [
                'nsslapd-db-abort-rate',
                'nsslapd-db-active-txns',
                'nsslapd-db-cache-hit',
                'nsslapd-db-cache-try',
                'nsslapd-db-cache-region-wait-rate',
                'nsslapd-db-cache-size-bytes',
                'nsslapd-db-clean-pages',
                'nsslapd-db-commit-rate',
                'nsslapd-db-deadlock-rate',
                'nsslapd-db-dirty-pages',
                'nsslapd-db-hash-buckets',
                'nsslapd-db-hash-elements-examine-rate',
                'nsslapd-db-hash-search-rate',
                'nsslapd-db-lock-conflicts',
                'nsslapd-db-lock-region-wait-rate',
                'nsslapd-db-lock-request-rate',
                'nsslapd-db-lockers',
                'nsslapd-db-configured-locks',
                'nsslapd-db-current-locks',
                'nsslapd-db-max-locks',
                'nsslapd-db-current-lock-objects',
                'nsslapd-db-max-lock-objects',
                'nsslapd-db-log-bytes-since-checkpoint',
                'nsslapd-db-log-region-wait-rate',
                'nsslapd-db-log-write-rate',
                'nsslapd-db-longest-chain-length',
                'nsslapd-db-page-create-rate',
                'nsslapd-db-page-read-rate',
                'nsslapd-db-page-ro-evict-rate',
                'nsslapd-db-page-rw-evict-rate',
                'nsslapd-db-page-trickle-rate',
                'nsslapd-db-page-write-rate',
                'nsslapd-db-pages-in-use',
                'nsslapd-db-txn-region-wait-rate',
           ]
        if db_lib == "mdb":
            self._backend_keys = [
                'dbenvmapmaxsize',
                'dbenvmapsize',
                'dbenvlastpageno',
                'dbenvlasttxnid',
                'dbenvmaxreaders',
                'dbenvnumreaders',
                'dbenvnumdbis',
                'waitingrwtxn',
                'activerwtxn',
                'abortrwtxn',
                'commitrwtxn',
                'granttimerwtxn',
                'lifetimerwtxn',
                'waitingrotxn',
                'activerotxn',
                'abortrotxn',
                'commitrotxn',
                'granttimerotxn',
                'lifetimerotxn',
           ]


    def get_status(self, use_json=False):
        if not self._backend_keys:
            self.__init2()
        return self.get_attrs_vals_utf8(self._backend_keys)


class MonitorBackend(DSLdapObject):
    """
    This is initialised from Backend in backend.py to get the right basedn.
    """

    def __init__(self, instance, dn=None):
        super(MonitorBackend, self).__init__(instance=instance, dn=dn)
        self._backend_keys = None

    def __init2(self):
        # Determine the key when really accessing the object with get_status
        # because config attrbute and connectio are not yet set in DirSrv
        # when __init is called
        db_lib = self._instance.get_db_lib()
        if db_lib == "bdb":
            self._backend_keys = [
                'readonly',
                'entrycachehits',
                'entrycachetries',
                'entrycachehitratio',
                'currententrycachesize',
                'maxentrycachesize',
                'currententrycachecount',
                'maxentrycachecount',
                'dncachehits',
                'dncachetries',
                'dncachehitratio',
                'currentdncachesize',
                'maxdncachesize',
                'currentdncachecount',
                'maxdncachecount',
            ]
            if ds_is_older("1.4.0", instance=self._instance):
                self._backend_keys.extend([
                    'normalizeddncachetries',
                    'normalizeddncachehits',
                    'normalizeddncachemisses',
                    'normalizeddncachehitratio',
                    'currentnormalizeddncachesize',
                    'maxnormalizeddncachesize',
                    'currentnormalizeddncachecount'
                ])
        if db_lib == "mdb":
            self._backend_keys = [
                'readonly',
                'entrycachehits',
                'entrycachetries',
                'entrycachehitratio',
                'currententrycachesize',
                'maxentrycachesize',
                'currententrycachecount',
                'maxentrycachecount',
            ]


    def get_status(self, use_json=False):
        if not self._backend_keys:
            self.__init2()
        result = {}
        all_attrs = self.get_all_attrs_utf8()
        for attr in self._backend_keys:
            result[attr] = all_attrs[attr]

        # Now gather all the dbfile* attributes
        for attr, val in all_attrs.items():
            # For bdb
            if attr.startswith('dbfile'):
                result[attr] = val
            # For lmdb
            if attr.startswith('dbi'):
                result[attr] = val

        return result


class MonitorChaining(DSLdapObject):
    """
    """
    def __init__(self, instance, dn=None):
        super(MonitorChaining, self).__init__(instance=instance, dn=dn)
        self._chaining_keys = [
            'nsaddcount',
            'nsdeletecount',
            'nsmodifycount',
            'nsrenamecount',
            'nssearchbasecount',
            'nssearchonelevelcount',
            'nssearchsubtreecount',
            'nsabandoncount',
            'nsbindcount',
            'nsunbindcount',
            'nscomparecount',
            'nsopenopconnectioncount',
            'nsopenbindconnectioncount'
        ]
        self._protected = False

    def get_status(self, use_json=False):
        return self.get_attrs_vals_utf8(self._chaining_keys)


class MonitorSNMP(DSLdapObject):
    """
    """
    def __init__(self, instance, dn=None):
        super(MonitorSNMP, self).__init__(instance=instance, dn=dn)
        self._dn = DN_MONITOR_SNMP
        self._snmp_keys = [
            'anonymousbinds',
            'unauthbinds',
            'simpleauthbinds',
            'strongauthbinds',
            'bindsecurityerrors',
            'inops',
            'readops',
            'compareops',
            'addentryops',
            'removeentryops',
            'modifyentryops',
            'modifyrdnops',
            'listops',
            'searchops',
            'onelevelsearchops',
            'wholesubtreesearchops',
            'referrals',
            'chainings',
            'securityerrors',
            'errors',
            'connections',
            'connectionseq',
            'connectionsinmaxthreads',
            'connectionsmaxthreadscount',
            'bytesrecv',
            'bytessent',
            'entriesreturned',
            'referralsreturned',
            'copyentries',
            'cacheentries',
            'cachehits',
        ]

    def get_status(self, use_json=False):
        return self.get_attrs_vals_utf8(self._snmp_keys)


class MonitorDiskSpace(DSLdapObject):
    """A class for representing "cn=disk space,cn=monitor" entry"""

    def __init__(self, instance, dn=None):
        super(MonitorDiskSpace, self).__init__(instance=instance, dn=dn)
        self._dn = "cn=disk space,cn=monitor"

    @classmethod
    def lint_uid(cls):
        return 'monitor-disk-space'

    def _lint_disk_space(self):
        threshold = 90
        if self._instance.status():
            partitions = self.get_attr_vals_utf8_l("dsDisk")
            for partition in partitions:
                parts = partition.split()
                percent = parts[4].split('=')[1].strip('"')
                if int(percent) >= threshold:
                    # this partition is over 90% full, not good
                    report = copy.deepcopy(DSDSLE0001)
                    report['detail'] = report['detail'].replace('PARTITION', parts[0].split('=')[1].strip('"'))
                    report['fix'] = report['fix'].replace('YOUR_INSTANCE', self._instance.serverid)
                    report['check'] = f'monitor-disk-space:disk_space'
                    yield report
        else:
            mounts = []
            paths = [self._instance.ds_paths.access_log,
                     self._instance.ds_paths.error_log,
                     self._instance.ds_paths.audit_log,
                     self._instance.ds_paths.security_log,
                     self._instance.ds_paths.config_dir,
                     self._instance.ds_paths.schema_dir,
                     self._instance.ds_paths.ldif_dir,
                     self._instance.ds_paths.inst_dir,
                     self._instance.ds_paths.db_dir]

            for path in paths:
                mount_point = get_mount_point(path)
                if mount_point not in mounts:
                    mounts.append(mount_point)

            for mount in mounts:
                disk_space = get_disk_space(mount)
                if disk_space['percent'] >= threshold:
                    # this partition is over 90% full, not good
                    report = copy.deepcopy(DSDSLE0001)
                    report['detail'] = report['detail'].replace('PARTITION', mount)
                    report['fix'] = report['fix'].replace('YOUR_INSTANCE', self._instance.serverid)
                    report['check'] = f'monitor-disk-space:disk_space'
                    yield report

    def get_disks(self):
        """Get an information about partitions which contains a Directory Server data"""

        return self.get_attr_vals_utf8_l("dsDisk")


class MonitorMemberOf(DSLdapObject):
    """A class for representing the "cn=MemberOf Plugin,cn=monitor" entry"""
    def __init__(self, instance, dn=None):
        super(MonitorMemberOf, self).__init__(instance=instance, dn=dn)
        self._dn = DN_MONITOR_MEMBEROF
        self._deferred_memberof_keys = [
            'CurrentTasks',
            'TotalAdded',
            'TotalRemoved',
            'TotalPending',
            'CompletionRate',
            'ThreadStatus',
            'QueueUtilisation',
        ]
        self._protected = False

    def get_status(self, use_json=False):
        try:
            return self.get_attrs_vals_utf8(self._deferred_memberof_keys)
        except ldap.NO_SUCH_OBJECT:
            raise ValueError("MemberOf monitoring not available (deferred processing disabled)")




# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2022 William Brown <william@blackhats.net.au>
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import json
import ldap
from lib389.plugins import (PAMPassThroughAuthPlugin,
                            PAMPassThroughAuthConfigs, PAMPassThroughAuthConfig)
from lib389.cli_base import CustomHelpFormatter
from lib389.cli_conf import add_generic_plugin_parsers, generic_object_edit, generic_object_add


arg_to_attr_pam = {
    'exclude_suffix': 'pamExcludeSuffix',
    'include_suffix': 'pamIncludeSuffix',
    'missing_suffix': 'pamMissingSuffix',
    'filter': 'pamFilter',
    'id_attr': 'pamIDAttr',
    'id_map_method': 'pamIDMapMethod',
    'fallback': 'pamFallback',
    'secure': 'pamSecure',
    'service': 'pamService'
}


def pam_pta_list(inst, basedn, log, args):
    log = log.getChild('pam_pta_list')
    configs = PAMPassThroughAuthConfigs(inst)
    result = []
    result_json = []
    for config in configs.list():
        if args.json:
            result_json.append(json.loads(config.get_all_attrs_json()))
        else:
            result.append(config.rdn)
    if args.json:
        log.info(json.dumps({"type": "list", "items": result_json}, indent=4))
    else:
        if len(result) > 0:
            for i in result:
                log.info(i)
        else:
            log.info("No PAM Pass Through Auth plugin config instances")


def pam_pta_add(inst, basedn, log, args):
    log = log.getChild('pam_pta_add')
    plugin = PAMPassThroughAuthPlugin(inst)
    props = {'cn': args.NAME}
    generic_object_add(PAMPassThroughAuthConfig, inst, log, args, arg_to_attr_pam, basedn=plugin.dn, props=props)


def pam_pta_edit(inst, basedn, log, args):
    log = log.getChild('pam_pta_edit')
    configs = PAMPassThroughAuthConfigs(inst)
    config = configs.get(args.NAME)
    generic_object_edit(config, log, args, arg_to_attr_pam)


def pam_pta_show(inst, basedn, log, args):
    log = log.getChild('pam_pta_show')
    configs = PAMPassThroughAuthConfigs(inst)
    config = configs.get(args.NAME)

    if not config.exists():
        raise ldap.NO_SUCH_OBJECT("Entry %s doesn't exists" % args.name)
    if args and args.json:
        o_str = config.get_all_attrs_json()
        log.info(o_str)
    else:
        log.info(config.display())


def pam_pta_del(inst, basedn, log, args):
    log = log.getChild('pam_pta_del')
    configs = PAMPassThroughAuthConfigs(inst)
    config = configs.get(args.NAME)
    config.delete()
    log.info("Successfully deleted the %s", config.dn)


def _add_parser_args_pam(parser):
    parser.add_argument('--exclude-suffix', nargs='+',
                        help='Specifies a suffix to exclude from PAM authentication (pamExcludeSuffix)')
    parser.add_argument('--include-suffix', nargs='+',
                        help='Sets a suffix to include for PAM authentication (pamIncludeSuffix)')
    parser.add_argument('--missing-suffix', choices=['ERROR', 'ALLOW', 'IGNORE', 'delete', ''],
                        help='Identifies how to handle missing include or exclude suffixes (pamMissingSuffix)')
    parser.add_argument('--filter',
                        help='Sets an LDAP filter to use to identify specific entries within '
                             'the included suffixes for which to use PAM pass-through authentication (pamFilter)')
    parser.add_argument('--id-attr',
                        help='Contains the attribute name which is used to hold the PAM user ID (pamIDAttr)')
    parser.add_argument('--id_map_method',
                        help='Sets the method to use to map the LDAP bind DN to a PAM identity (pamIDMapMethod)')
    parser.add_argument('--fallback', choices=['TRUE', 'FALSE'], type=str.upper,
                        help='Sets whether to fallback to regular LDAP authentication '
                             'if PAM authentication fails (pamFallback)')
    parser.add_argument('--secure', choices=['TRUE', 'FALSE'], type=str.upper,
                        help='Requires secure TLS connection for PAM authentication (pamSecure)')
    parser.add_argument('--service',
                        help='Contains the service name to pass to PAM (pamService)')


def create_parser(subparsers):
    passthroughauth_parser = subparsers.add_parser('pam-pass-through-auth',
                                                   help='Manage and configure Pass-Through Authentication plugins '
                                                        '(LDAP URLs and PAM)')
    subcommands = passthroughauth_parser.add_subparsers(help='action')

    add_generic_plugin_parsers(subcommands, PAMPassThroughAuthPlugin)

    list_pam = subcommands.add_parser('list', help='Lists PAM configurations', formatter_class=CustomHelpFormatter)
    list_pam.set_defaults(func=pam_pta_list)

    pam = subcommands.add_parser('config', help='Manage PAM PTA configurations.', formatter_class=CustomHelpFormatter)
    pam.add_argument('NAME', help='The PAM PTA configuration name')
    subcommands_pam = pam.add_subparsers(help='action')

    add = subcommands_pam.add_parser('add', help='Add the config entry', formatter_class=CustomHelpFormatter)
    add.set_defaults(func=pam_pta_add)
    _add_parser_args_pam(add)
    edit = subcommands_pam.add_parser('set', help='Edit the config entry', formatter_class=CustomHelpFormatter)
    edit.set_defaults(func=pam_pta_edit)
    _add_parser_args_pam(edit)
    show = subcommands_pam.add_parser('show', help='Display the config entry', formatter_class=CustomHelpFormatter)
    show.set_defaults(func=pam_pta_show)
    delete = subcommands_pam.add_parser('delete', help='Delete the config entry', formatter_class=CustomHelpFormatter)
    delete.set_defaults(func=pam_pta_del)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2020 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import json
from lib389.saslmap import SaslMapping, SaslMappings
from lib389.utils import ensure_str
from lib389.cli_base import (
    populate_attr_arguments,
    _generic_list,
    _generic_get,
    _generic_create,
    _generic_delete,
    _get_arg,
    _get_attributes,
    _warn,
    CustomHelpFormatter
    )

SINGULAR = SaslMapping
MANY = SaslMappings
RDN = 'cn'


def sasl_map_list(inst, basedn, log, args):
    if args.details:
        # List SASL mappings with details
        mappings = SaslMappings(inst).list()
        result = {"type": "list", "items": []}
        for sasl_map in mappings:
            if args.json:
                entry = sasl_map.get_all_attrs_json()
                # Append decoded json object, because we are going to dump it later
                result['items'].append(json.loads(entry))
            else:
                log.info(sasl_map.display())
        if args.json:
            log.info(json.dumps(result, indent=4))
    else:
        _generic_list(inst, basedn, log.getChild('sasl_map_list'), MANY, args)


def sasl_map_get(inst, basedn, log, args):
    rdn = _get_arg(args.selector, msg="Enter %s to retrieve" % RDN)
    _generic_get(inst, basedn, log.getChild('sasl_map_get'), MANY, rdn, args)


def sasl_map_create(inst, basedn, log, args):
    kwargs = _get_attributes(args, SaslMapping._must_attributes)
    if kwargs['nssaslmappriority'] == '':
        kwargs['nssaslmappriority'] = '100'  # Default
    _generic_create(inst, basedn, log.getChild('sasl_map_create'), MANY, kwargs, args)


def sasl_map_delete(inst, basedn, log, args, warn=True):
    found = False
    mappings = SaslMappings(inst).list()
    map_name = args.map_name.lower()

    for saslmap in mappings:
        cn = ensure_str(saslmap.get_attr_val('cn')).lower()
        if cn == map_name:
            dn = saslmap.dn
            found = True
            break
    if not found:
        raise ValueError("Unable to find a SASL mapping with the name: ({})".format(args.map_name))

    if warn and args.json is False:
        _warn(dn, msg="Deleting %s %s" % (SINGULAR.__name__, dn))
    _generic_delete(inst, basedn, log.getChild('sasl_map_delete'), SINGULAR, dn, args)

def sasl_get_supported(inst, basedn, log, args):
    """Get a list of the supported sasl mechanisms"""
    mechs = inst.rootdse.supported_sasl()
    if args.json:
        result = {'type': 'list', 'items': mechs}
        log.info(json.dumps(result, indent=4, ))
    else:
        for mech in mechs:
            log.info(mech)


def sasl_get_available(inst, basedn, log, args):
    """Get a list of the available sasl mechanisms"""
    mechs = inst.rootdse.available_sasl()
    if args.json:
        result = {'type': 'list', 'items': mechs}
        log.info(json.dumps(result, indent=4, ))
    else:
        for mech in mechs:
            log.info(mech)


def create_parser(subparsers):
    sasl_parser = subparsers.add_parser('sasl', help='Manage SASL mappings', formatter_class=CustomHelpFormatter)

    subcommands = sasl_parser.add_subparsers(help='sasl')

    list_mappings_parser = subcommands.add_parser('list', help='Display available SASL mappings', formatter_class=CustomHelpFormatter)
    list_mappings_parser.set_defaults(func=sasl_map_list)
    list_mappings_parser.add_argument('--details', action='store_true', default=False,
        help="Displays each SASL mapping in detail")

    get_mech_parser= subcommands.add_parser('get-mechs', help='Display the SASL mechanisms that the server will accept', formatter_class=CustomHelpFormatter)
    get_mech_parser.set_defaults(func=sasl_get_supported)

    get_mech_parser= subcommands.add_parser('get-available-mechs', help='Display the SASL mechanisms that are available to the server', formatter_class=CustomHelpFormatter)
    get_mech_parser.set_defaults(func=sasl_get_available)

    get_parser = subcommands.add_parser('get', help='Displays SASL mappings', formatter_class=CustomHelpFormatter)
    get_parser.set_defaults(func=sasl_map_get)
    get_parser.add_argument('selector', nargs='?', help='The SASL mapping name to display')

    create_parser = subcommands.add_parser('create', help='Create a SASL mapping ', formatter_class=CustomHelpFormatter)
    create_parser.set_defaults(func=sasl_map_create)
    populate_attr_arguments(create_parser, SaslMapping._must_attributes)

    delete_parser = subcommands.add_parser('delete', help='Deletes the SASL object', formatter_class=CustomHelpFormatter)
    delete_parser.set_defaults(func=sasl_map_delete)
    delete_parser.add_argument('map_name', help='The SASL mapping name ("cn" value)')




# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2020 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---


from lib389._mapped_object import DSLdapObjects, DSLdapObject


class EncryptedAttr(DSLdapObject):
    """Encrypted Attribute DSLdapObject with:
    - must attributes = ['cn', 'nsEncryptionAlgorithm']
    - RDN attribute is 'cn'

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param dn: Encrypted Attribute DN
    :type dn: str
    """

    def __init__(self, instance, dn=None):
        super(EncryptedAttr, self).__init__(instance, dn)
        self._rdn_attribute = 'cn'
        self._must_attributes = ['cn', 'nsEncryptionAlgorithm']
        self._create_objectclasses = ['top', 'nsAttributeEncryption']
        self._protected = False


class EncryptedAttrs(DSLdapObjects):
    """DSLdapObjects that represents Encrypted Attribute

    :param instance: An instance
    :type instance: lib389.DirSrv
    :param basedn: DN of suffix container.
    :type basedn: str
    """

    def __init__(self, instance, basedn):
        super(EncryptedAttrs, self).__init__(instance=instance)
        self._objectclasses = ['nsAttributeEncryption']
        self._filterattrs = ['cn']
        self._childobject = EncryptedAttr
        self._basedn = basedn



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2015 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
#
from lib389._constants import *
from lib389.utils import ensure_bytes, ensure_str
from lib389 import DirSrv, Entry
import pytest
import time
import shutil
import datetime as dt

INSTANCE_PORT = 54321
INSTANCE_SERVERID = 'standalone'


class TopologyStandalone(object):
    def __init__(self, standalone):
        standalone.open()
        self.standalone = standalone


@pytest.fixture(scope="module")
def topology(request):
    standalone = DirSrv(verbose=False)
    standalone.log.debug("Instance allocated")
    args = {
            SER_PORT: INSTANCE_PORT,
            # SER_DEPLOYED_DIR:  INSTANCE_PREFIX,
            SER_SERVERID_PROP: INSTANCE_SERVERID}
    standalone.allocate(args)
    if standalone.exists():
        standalone.delete()
    standalone.create()
    standalone.open()
    standalone.config.set('nsslapd-accesslog-logbuffering', 'off')

    def fin():
        standalone.delete()
    request.addfinalizer(fin)
    # We have to wait for time to elapse for the access log to be flushed.

    return TopologyStandalone(standalone)


def test_access_log_rotation(topology):
    """
    Check we can parse rotated logs as well as active log.
    """

    # Artificially rotate the log.
    lpath = topology.standalone.ds_access_log._get_log_path()
    shutil.copyfile(lpath, lpath + ensure_bytes('.20160515-104822'))
    # check we have the right number of lines.
    access_lines = topology.standalone.ds_access_log.readlines_archive()
    assert(len(access_lines) > 0)
    access_lines = topology.standalone.ds_access_log.match_archive('.*fd=.*')
    assert(len(access_lines) > 0)


def test_access_log(topology):
    """Check the parsing of the access log"""
    access_lines = topology.standalone.ds_access_log.readlines()
    assert(len(access_lines) > 0)
    access_lines = topology.standalone.ds_access_log.match('.*fd=.*')
    assert(len(access_lines) > 0)
    # Test the line parser in a basic way.
    assert(
        topology.standalone.ds_access_log.parse_line('[27/Apr/2016:12:49:49.726093186 +1000] conn=1 fd=64 slot=64 connection from ::1 to ::1') ==
        {
            'slot': '64', 'remote': '::1', 'action': 'CONNECT', 'timestamp': '[27/Apr/2016:12:49:49.726093186 +1000]', 'fd': '64', 'conn': '1', 'local': '::1',
            'datetime': dt.datetime(2016, 4, 27, 12, 49, 49, 726093, tzinfo=dt.timezone(dt.timedelta(seconds=36000)))
        }
    )
    assert(
        topology.standalone.ds_access_log.parse_line('[27/Apr/2016:12:49:49.727235997 +1000] conn=1 op=2 SRCH base="cn=config" scope=0 filter="(objectClass=*)" attrs="nsslapd-instancedir nsslapd-errorlog nsslapd-accesslog nsslapd-auditlog nsslapd-certdir nsslapd-schemadir nsslapd-bakdir nsslapd-ldifdir"') ==  # noqa
        {
            'rem': 'base="cn=config" scope=0 filter="(objectClass=*)" attrs="nsslapd-instancedir nsslapd-errorlog nsslapd-accesslog nsslapd-auditlog nsslapd-certdir nsslapd-schemadir nsslapd-bakdir nsslapd-ldifdir"',  # noqa
            'action': 'SRCH', 'timestamp': '[27/Apr/2016:12:49:49.727235997 +1000]', 'conn': '1', 'op': '2',
            'datetime': dt.datetime(2016, 4, 27, 12, 49, 49, 727235, tzinfo=dt.timezone(dt.timedelta(seconds=36000)))
        }
    )
    assert(
        topology.standalone.ds_access_log.parse_line('[27/Apr/2016:12:49:49.736297002 +1000] conn=1 op=4 fd=64 closed - U1') ==
        {
            'status': 'U1', 'fd': '64', 'action': 'DISCONNECT', 'timestamp': '[27/Apr/2016:12:49:49.736297002 +1000]', 'conn': '1', 'op': '4',
            'datetime': dt.datetime(2016, 4, 27, 12, 49, 49, 736297, tzinfo=dt.timezone(dt.timedelta(seconds=36000)))
        }
    )
    assert(
        topology.standalone.ds_access_log.parse_line('[27/Apr/2016:12:49:49.736297002 -1000] conn=1 op=4 fd=64 closed - U1') ==
        {
            'status': 'U1', 'fd': '64', 'action': 'DISCONNECT', 'timestamp': '[27/Apr/2016:12:49:49.736297002 -1000]', 'conn': '1', 'op': '4',
            'datetime': dt.datetime(2016, 4, 27, 12, 49, 49, 736297, tzinfo=dt.timezone(dt.timedelta(seconds=-36000)))
        }
    )


def test_error_log(topology):
    """Check the parsing of the error log"""
    # No need to sleep, it's not buffered.
    error_lines = topology.standalone.ds_error_log.readlines()
    assert(len(error_lines) > 0)
    error_lines = topology.standalone.ds_error_log.match('.*started.*')
    assert(len(error_lines) > 0)

    assert(
        topology.standalone.ds_error_log.parse_line('[27/Apr/2016:13:46:35.775670167 +1000] slapd started.  Listening on All Interfaces port 54321 for LDAP requests') ==  # noqa
        {
            'timestamp': '[27/Apr/2016:13:46:35.775670167 +1000]', 'message': 'slapd started.  Listening on All Interfaces port 54321 for LDAP requests',
            'datetime': dt.datetime(2016, 4, 27, 13, 46, 35, 775670, tzinfo=dt.timezone(dt.timedelta(seconds=36000)))
        }
    )


if __name__ == "__main__":
    CURRENT_FILE = os.path.realpath(__file__)
    pytest.main("-s -vv %s" % CURRENT_FILE)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2017 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

from lib389.utils import ds_is_newer
from lib389._constants import INSTALL_LATEST_CONFIG
from .config_001003006 import c001003006, c001003006_sample_entries
from .config_001004000 import c001004000, c001004000_sample_entries
from .config_001004002 import c001004002, c001004002_sample_entries
from .config_002003000 import c002003000, c002003000_sample_entries


def get_config(version):
    # We do this to avoid test breaking on older version that may
    # not expect the new default layout.
    if (version == INSTALL_LATEST_CONFIG and ds_is_newer('2.3.0')):
        return c002003000
    if (version == INSTALL_LATEST_CONFIG and ds_is_newer('1.4.2')):
        return c001004002
    elif (version == INSTALL_LATEST_CONFIG and ds_is_newer('1.4.0')):
        return c001004000
    elif (version == INSTALL_LATEST_CONFIG):
        return c001003006
    elif (version == '002003000' and ds_is_newer('2.3.0')):
        return c002003000
    elif (version == '001004002' and ds_is_newer('1.4.2')):
        return c001004002
    elif (version == '001004000' and ds_is_newer('1.4.0')):
        return c001004000
    elif (version == '001003006'):
        return c001003006
    raise Exception('version %s no match' % version)


def get_sample_entries(version):
    if (version == INSTALL_LATEST_CONFIG):
        return c002003000_sample_entries
    elif (version == '002003000'):
        return c002003000_sample_entries
    elif (version == '001004002'):
        return c001004002_sample_entries
    elif (version == '001004000'):
        return c001004000_sample_entries
    elif (version == '001003006'):
        return c001003006_sample_entries
    raise Exception('version %s no match' % version)




# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2015 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import string
import os
import sys
import pytest
from tempfile import mkdtemp
from shutil import rmtree

MAJOR, MINOR, _, _, _ = sys.version_info


def ldrop(numchar, inputstr):
    """
        Drop numchar characters from the beginning of each line in inputstring.
    """
    if MAJOR >= 3:
        return "\n".join(map(lambda x: x[numchar:], inputstr.split("\n")))
    else:
        return "\n".join([x[numchar:] for x in string.split(inputstr, "\n")])


@pytest.fixture
def fake_ds_class():
    """
        Returns fake DSInstance-like class
    """
    class FakeDSInstance(object):
        def return_one(self):
            return 1

    return FakeDSInstance


@pytest.fixture(scope="module")
def fake_ds_modules(request):
    """
        Returns path to directory containing fake modules.
    """

    # module_dir contains fake ds modules, that can be used for testing
    module_dir = mkdtemp()
    old_cwd = os.getcwd()

    # when fixture is destroyed, we change back to old cwd
    # and delete fake module dir
    def fin():
        os.chdir(old_cwd)
        rmtree(module_dir)
    request.addfinalizer(fin)

    # make module_dir a python package
    with open(module_dir + "/__init__.py", 'a'):
        os.utime(module_dir + "/__init__.py", None)

    # create module module_dir/repl.py with few functions
    with open(module_dir + "/repl.py", "w") as repl_file:
        repl_module = """
            def with_no_arg(ds):
                return ds

            def with_one_arg(ds, arg):
                return ds, arg

            def return_argument_given(ds, arg):
                return arg

            def with_kwarg(ds, kwarg=2):
                return kwarg

            def with_all(ds, arg1, arg2, kwarg1=1, kwarg2=2):
                return arg1, arg2, kwarg1, kwarg2

            """
        # Drop 12 chars from left and write to file
        repl_file.write(ldrop(12, repl_module))

    os.mkdir(module_dir + "/plugin")

    # make module_dir/plugin a python package
    with open(module_dir + "/plugin/__init__.py", 'a'):
        os.utime(module_dir + "/plugin/__init__.py", None)

    # create module module_dir/plugin/automember.py with few functions
    with open(module_dir + "/plugin/automember.py", "w") as automem_file:
        automember_module = """
            def with_no_arg(ds):
                return ds

            def with_one_arg(ds, arg):
                return ds, arg

            def calling_ds_method(ds, arg):
                return (ds, ds.return_one(), arg)

            def calling_repl(ds, arg1, arg2):
                return ds.repl.return_argument_given(1) + arg1 + arg2

            def calling_repl_with_kwarg2(ds):
                return ds.repl.with_kwarg()

            def calling_repl_with_kwarg1(ds):
                return ds.repl.with_kwarg(kwarg=1)

            def with_all(ds, arg1, arg2, kwarg1=1, kwarg2=2):
                a, b, c, d = ds.repl.with_all(1, 4, kwarg1=3)
                return a + arg1, b + arg2, c + kwarg1, d + kwarg2

            """
        # Drop 12 chars from left and write to file
        automem_file.write(ldrop(12, automember_module))

    os.chdir(os.path.dirname(module_dir))
    sys.path.append(os.path.dirname(module_dir))
    return module_dir



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2021 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---
from lib389.plugins import RetroChangelogPlugin
from lib389.cli_conf import add_generic_plugin_parsers, generic_object_edit, generic_object_add_attr, generic_object_del_attr
from lib389.cli_base import CustomHelpFormatter
from lib389.utils import validate_max_age


arg_to_attr = {
    'is_replicated': 'isReplicated',
    'attribute': 'nsslapd-attribute',
    'directory': 'nsslapd-changelogdir',
    'max_age': 'nsslapd-changelogmaxage',
    'trim_interval': 'nsslapd-changelog-trim-interval',
    'exclude_suffix': 'nsslapd-exclude-suffix',
    'exclude_attrs': 'nsslapd-exclude-attrs',
}

def retrochangelog_edit(inst, basedn, log, args):
    log = log.getChild('retrochangelog_edit')
    if args.max_age:
        validate_max_age(args.max_age, ignore_value="0")
    plugin = RetroChangelogPlugin(inst)
    generic_object_edit(plugin, log, args, arg_to_attr)


def retrochangelog_add(inst, basedn, log, args):
    log = log.getChild('retrochangelog_add')
    if args.max_age:
        validate_max_age(args.max_age, ignore_value="0")
    plugin = RetroChangelogPlugin(inst)
    generic_object_add_attr(plugin, log, args, arg_to_attr)

def retrochangelog_del(inst, basedn, log, args):
    log = log.getChild('retrochangelog_add')
    plugin = RetroChangelogPlugin(inst)
    generic_object_del_attr(plugin, log, args, arg_to_attr)

def _add_parser_args(parser):
    parser.add_argument('--is-replicated', choices=['TRUE', 'FALSE'], type=str.upper,
                        help='Sets a flag to indicate on a change in the changelog whether the change is newly made '
                             'on that server or whether it was replicated over from another server (isReplicated)')
    parser.add_argument('--attribute',
                        help='Specifies another Directory Server attribute which must be included in '
                             'the retro changelog entries (nsslapd-attribute)')
    parser.add_argument('--directory',
                        help='Specifies the name of the directory in which the changelog database '
                             'is created the first time the plug-in is run')
    parser.add_argument('--max-age',
                        help='Specifies the maximum age of any entry in the changelog. '
                             'The value must be a number followed by a duration unit [sSmMhHdDwW]. '
                             'Used to trim the changelog (nsslapd-changelogmaxage)')
    parser.add_argument('--trim-interval',
                        help='. nsslapd-changelog-trim-interval)')
    parser.add_argument('--exclude-suffix', nargs='*',
                        help='Specifies the suffix which will be excluded from the scope of the plugin '
                             '(nsslapd-exclude-suffix)')
    parser.add_argument('--exclude-attrs', nargs='*',
                        help='Specifies the attributes which will be excluded from the scope of the plugin '
                             '(nsslapd-exclude-attrs)')


def create_parser(subparsers):
    retrochangelog = subparsers.add_parser('retro-changelog', help='Manage and configure Retro Changelog plugin', formatter_class=CustomHelpFormatter)
    subcommands = retrochangelog.add_subparsers(help='action')
    add_generic_plugin_parsers(subcommands, RetroChangelogPlugin)

    edit = subcommands.add_parser('set', help='Edit the plugin', formatter_class=CustomHelpFormatter)
    edit.set_defaults(func=retrochangelog_edit)
    _add_parser_args(edit)

    addp = subcommands.add_parser('add', help='Add attributes to the plugin', formatter_class=CustomHelpFormatter)
    addp.set_defaults(func=retrochangelog_add)
    _add_parser_args(addp)

    delp = subcommands.add_parser('del', help='Delete an attribute from plugin scope', formatter_class=CustomHelpFormatter)
    delp.set_defaults(func=retrochangelog_del)
    _add_parser_args(delp)



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2020 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

from ldap import dn

from lib389.idm.domain import Domain
from lib389.idm.organization import Organization
from lib389.idm.organizationalunit import OrganizationalUnit
from lib389.idm.nscontainer import nsContainer
from lib389.idm.country import Country
from lib389.utils import ensure_str


def create_base_domain(instance, basedn):
    """Create the base domain object"""

    domain = Domain(instance, dn=basedn)
    # Explode the dn to get the first bit.
    avas = dn.str2dn(basedn)
    dc_ava = avas[0][0][1]

    domain.create(properties={
        # I think in python 2 this forces unicode return ...
        'dc': dc_ava,
        'description': basedn,
    })
    # ACI can be added later according to your needs

    return domain


def create_base_org(instance, basedn):
    """Create the base organization object"""

    org = Organization(instance, dn=basedn)
    # Explode the dn to get the first bit.
    avas = dn.str2dn(basedn)
    o_ava = avas[0][0][1]

    org.create(properties={
        # I think in python 2 this forces unicode return ...
        'o': o_ava,
        'description': basedn,
    })

    return org


def create_base_orgunit(instance, basedn):
    """Create the base org unit object for a org unit"""

    orgunit = OrganizationalUnit(instance, dn=basedn)
    # Explode the dn to get the first bit.
    avas = dn.str2dn(basedn)
    ou_ava = avas[0][0][1]

    orgunit.create(properties={
        # I think in python 2 this forces unicode return ...
        'ou': ou_ava,
        'description': basedn,
    })

    return orgunit


def create_base_cn(instance, basedn):
    """Create the base nsContainer object"""

    cn = nsContainer(instance, dn=basedn)
    # Explode the dn to get the first bit.
    avas = dn.str2dn(basedn)
    cn_ava = avas[0][0][1]

    cn.create(properties={
        # I think in python 2 this forces unicode return ...
        'cn': cn_ava,
    })

    return cn

def create_base_c(instance, basedn):
    """Create the base country object"""

    c = Country(instance, dn=basedn)
    # Explode the dn to get the first bit.
    avas = dn.str2dn(basedn)
    c_ava = avas[0][0][1]

    c.create(properties={
        'c': c_ava,
    })

    return c


class sampleentries(object):
    def __init__(self, instance, basedn):
        self._instance = instance
        self._basedn = ensure_str(basedn)
        self.description = None
        self.version = None

    def apply(self):
        self._apply()

    def _configure_base(self, add_acis=True):
        suffix_rdn_attr = self._basedn.split('=')[0].lower()
        suffix_obj = None
        if suffix_rdn_attr == 'dc':
            suffix_obj = create_base_domain(self._instance, self._basedn)
            aci_vals = ['dc', 'domain']
        elif suffix_rdn_attr == 'c':
            suffix_obj = create_base_c(self._instance, self._basedn)
            aci_vals = ['c', 'country']
        elif suffix_rdn_attr == 'o':
            suffix_obj = create_base_org(self._instance, self._basedn)
            aci_vals = ['o', 'organization']
        elif suffix_rdn_attr == 'ou':
            suffix_obj = create_base_orgunit(self._instance, self._basedn)
            aci_vals = ['ou', 'organizationalunit']
        elif suffix_rdn_attr == 'cn':
            suffix_obj = create_base_cn(self._instance, self._basedn)
            aci_vals = ['cn', 'nscontainer']
        else:
            # Unsupported rdn
            raise ValueError("Suffix RDN '{}' in '{}' is not supported.  Supported RDN's are: 'c', 'cn', 'dc', 'o', and 'ou'".format(suffix_rdn_attr, self._basedn))

        if add_acis:
            suffix_obj.add('aci', [
                # Allow reading the base domain object
                '(targetattr="' + aci_vals[0] + ' || description || objectClass")(targetfilter="(objectClass=' + aci_vals[1] + ')")(version 3.0; acl "Enable anyone ' + aci_vals[1] + ' read"; allow (read, search, compare)(userdn="ldap:///anyone");)',
                # Allow reading the ou
                '(targetattr="ou || objectClass")(targetfilter="(objectClass=organizationalUnit)")(version 3.0; acl "Enable anyone ou read"; allow (read, search, compare)(userdn="ldap:///anyone");)'
            ])
        return suffix_obj

    def _apply(self):
        raise Exception('Not implemented')





# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2015 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

"""Aci class to help parse and create ACIs.

You will access this via the Entry Class.
"""

import ldap

# Helpers to detect common patterns in aci
def _aci_any_targetattr_ne(aci):
    """Returns True if any of the targetattr types is a != type"""

    potential = False
    if 'targetattr' in aci.acidata:
        for ta in aci.acidata['targetattr']:
            if ta['equal'] is False:
                # Got, it lets do this
                potential = True
    return potential


class Aci(object):
    """An object that helps to work with agreement entry

    :param conn: An instance
    :type conn: lib389.DirSrv
    """

    def __init__(self, conn):
        self.conn = conn
        self.log = conn.log

    def list(self, basedn, scope=ldap.SCOPE_SUBTREE):
        """List all acis in the directory server below the basedn confined by
        scope.

        :param basedn: Base DN
        :type basedn: str
        :param scope: ldap.SCOPE_SUBTREE, ldap.SCOPE_BASE,
                       ldap.SCOPE_ONELEVEL, ldap.SCOPE_SUBORDINATE
        :type scope: int

        :returns: A list of EntryAci objects
        """

        acis = []
        rawacientries = self.conn.search_s(basedn, scope, 'aci=*', ['aci'])
        for rawacientry in rawacientries:
            acis += rawacientry.getAcis()
        return acis

    def lint(self, basedn, scope=ldap.SCOPE_SUBTREE):
        """Validate and check for potential aci issues.

        Given a scope and basedn, this will retrieve all the aci's below.
        A number of checks are then run on the aci in isolation, and
        in groups.

        :param basedn: Base DN
        :type basedn: str
        :param scope: ldap.SCOPE_SUBTREE, ldap.SCOPE_BASE,
                       ldap.SCOPE_ONELEVEL, ldap.SCOPE_SUBORDINATE
        :type scope: int

        :returns: A tuple of (bool, list( dict ))
                   - Bool represents if the acis pass or fail as a whole.
                   - The list contains a list of warnings about your acis.
                   - The dict is structured as::

                       {
                         name: "" # DSALEXXXX
                         severity: "" # LOW MEDIUM HIGH
                         detail: "" # explination
                       }
        """

        result = True
        # Not thread safe!!!
        self.warnings = []

        acis = self.list(basedn, scope)
        # Checks again "all acis" go here.
        self._lint_dsale_0001_ne_internal(acis)
        self._lint_dsale_0002_ne_mult_subtree(acis)
        # checks again individual here

        if len(self.warnings) > 0:
            result = False

        return (result, self.warnings)

    def format_lint(self, warnings):
        """Takes the array of warnings and returns a formatted string.

        :param warnings: The array of warnings
        :type warnings: dict

        :returns: Formatted string or warnings
        """

        buf = "-------------------------------------------------------------------------------"

        for warning in warnings:
            buf += """
Directory Server Aci Lint Error: {DSALE}
Severity: {SEVERITY}

Affected Acis:
{ACIS}

Details: {DETAIL}

Advice: {FIX}
-------------------------------------------------------------------------------""".format(
                DSALE=warning['dsale'],
                SEVERITY=warning['severity'],
                ACIS=warning['acis'],
                DETAIL=warning['detail'],
                FIX=warning['fix'],
            )
        return buf

    # These are the aci lint checks.

    def _lint_dsale_0001_ne_internal(self, acis):
        """Check for the presence of "not equals" attributes that will inadvertantly
        allow the return / modification of internal attributes.
        """

        affected = []
        for aci in acis:
            if 'targetattr' in aci.acidata:
                for ta in aci.acidata['targetattr']:
                    if ta['equal'] is False:
                        affected.append(aci.acidata['rawaci'])

        if len(affected) > 0:
            self.warnings.append(
                {
                    'dsale': 'DSALE0001',
                    'severity': 'HIGH',
                    'acis': "\n".join(affected),
                    'detail': """
An aci of the form "(targetAttr!="attr")" exists on your system. This aci
will internally be expanded to mean "all possible attributes including system,
excluding the listed attributes".

This may allow access to a bound user or anonymous to read more data about
directory internals, including aci state or user limits. In the case of write
acis it may allow a dn to set their own resource limits, unlock passwords or
their own aci.

The ability to change the aci on the object may lead to privilege escalation in
some cases.
                    """,
                    'fix': """
Convert the aci to the form "(targetAttr="x || y || z")".
                    """
                }
            )

    def _lint_dsale_0002_ne_mult_subtree(self, acis):
        """This check will show pairs or more of aci that match the same subtree
        with a != rute. These can cause the other rule to be invalidated!
        """

        affected = []
        for aci in acis:
            # The aci has to be a NE, else don't bother checking
            if not _aci_any_targetattr_ne(aci):
                continue

            affect = False
            buf = "%s %s\n" % (aci.entry.dn, aci.acidata['rawaci'])
            for aci_inner in acis:

                if aci_inner == aci:
                    # Don't compare to self!
                    continue
                # Check the inner is a not equal also
                if not _aci_any_targetattr_ne(aci_inner):
                    continue

                # Check if the dn is a substring, ie child, or equal.
                if aci.entry.dn.endswith(aci_inner.entry.dn):
                    # alias the allow rules
                    allow_inner = set(aci_inner.acidata['allow'][0]['values'])
                    allow_outer = set(aci.acidata['allow'][0]['values'])
                    if len(allow_inner & allow_outer):
                        buf += "|- %s %s\n" % (aci_inner.entry.dn, aci_inner.acidata['rawaci'])
                        affect = True

            if affect:
                affected.append(buf)

        if len(affected) > 0:
            self.warnings.append(
                {
                    'dsale': 'DSALE0002',
                    'severity': 'HIGH',
                    'acis': "\n".join(affected),
                    'detail': """
Acis on your system exist which are both not equals targetattr, and overlap in
scope.

The way that directory server processes these, is to invert them to to white
lists, then union the results.

As a result, these acis *may* allow access to the attributes you want them to
exclude.

Consider:

aci: (targetattr !="cn")(version 3.0;acl "Self write all but cn";allow (write)
    (userdn = "ldap:///self");)
aci: (targetattr !="sn")(version 3.0;acl "Self write all but sn";allow (write)
    (userdn = "ldap:///self");)

This combination allows self write to *all* attributes within the subtree.

In cases where the target is members of a group, it may allow a member who is
within two groups to have elevated privilege.
                    """,
                    'fix': """
Convert the aci to the form "(targetAttr="x || y || z")".

Prevent the acis from overlapping, and have them on unique subtrees.
                    """
                }
            )



# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2017 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

from .config import baseconfig, configoperation
from .sample import sampleentries, create_base_domain

from lib389.idm.organizationalunit import OrganizationalUnits
from lib389.idm.group import UniqueGroups, UniqueGroup

from lib389.plugins import WhoamiPlugin

class c001003006_sample_entries(sampleentries):
    def __init__(self, instance, basedn):
        super(c001003006_sample_entries, self).__init__(instance, basedn)
        self.description = """Apply sample entries matching the 1.3.6 sample data and access controls."""
        self.version = "c001003006"

    # All the checks are done, apply them.
    def _apply(self):
        # Create the base domain object
        domain = create_base_domain(self._instance, self._basedn)
        domain.add('aci' , '(targetattr ="*")(version 3.0;acl "Directory Administrators Group";allow (all) (groupdn = "ldap:///cn=Directory Administrators,{BASEDN}");)'.format(BASEDN=self._basedn))

        # Create the OUs
        ous = OrganizationalUnits(self._instance, self._basedn)
        ous.create(properties = {
            'ou': 'Groups',
        })
        ous.create(properties = {
            'ou': 'People',
            'aci' : [
                '(targetattr ="userpassword || telephonenumber || facsimiletelephonenumber")(version 3.0;acl "Allow self entry modification";allow (write)(userdn = "ldap:///self");)',
                '(targetattr !="cn || sn || uid")(targetfilter ="(ou=Accounting)")(version 3.0;acl "Accounting Managers Group Permissions";allow (write)(groupdn = "ldap:///cn=Accounting Managers,ou=groups,{BASEDN}");)'.format(BASEDN=self._basedn),
                '(targetattr !="cn || sn || uid")(targetfilter ="(ou=Human Resources)")(version 3.0;acl "HR Group Permissions";allow (write)(groupdn = "ldap:///cn=HR Managers,ou=groups,{BASEDN}");)'.format(BASEDN=self._basedn),
                '(targetattr !="cn ||sn || uid")(targetfilter ="(ou=Product Testing)")(version 3.0;acl "QA Group Permissions";allow (write)(groupdn = "ldap:///cn=QA Managers,ou=groups,{BASEDN}");)'.format(BASEDN=self._basedn),
                '(targetattr !="cn || sn || uid")(targetfilter ="(ou=Product Development)")(version 3.0;acl "Engineering Group Permissions";allow (write)(groupdn = "ldap:///cn=PD Managers,ou=groups,{BASEDN}");)'.format(BASEDN=self._basedn),
            ]
        })
        ous.create(properties = {
            'ou': 'Special Users',
            'description' : 'Special Administrative Accounts',
        })
        # Create the groups.
        ugs = UniqueGroups(self._instance, self._basedn)
        ugs.create(properties = {
            'cn': 'Accounting Managers',
            'description': 'People who can manage accounting entries',
            'ou': 'groups',
            'uniqueMember' : self._instance.binddn,
        })
        ugs.create(properties = {
            'cn': 'HR Managers',
            'description': 'People who can manage HR entries',
            'ou': 'groups',
            'uniqueMember' : self._instance.binddn,
        })
        ugs.create(properties = {
            'cn': 'QA Managers',
            'description': 'People who can manage QA entries',
            'ou': 'groups',
            'uniqueMember' : self._instance.binddn,
        })
        ugs.create(properties = {
            'cn': 'PD Managers',
            'description': 'People who can manage engineer entries',
            'ou': 'groups',
            'uniqueMember' : self._instance.binddn,
        })
        # Create the directory Admin group.
        # We can't use the group factory here, as we need a custom DN override.
        da_ug = UniqueGroup(self._instance)
        da_ug._dn = 'cn=Directory Administrators,%s' % self._basedn
        da_ug.create(properties={
            'cn': 'Directory Administrators',
            'uniqueMember' : self._instance.binddn,
        })
        # DONE!

### Operations to be filled in soon!

class c001003006_whoami_plugin(configoperation):
    def __init__(self, instance):
        super(c001003006_whoami_plugin, self).__init__(instance)
        self.install = True
        self.upgrade = True
        self.description = "Enable the Whoami Plugin"

    def _apply(self):
        # This needs to change from create to "ensure"?
        whoami_plugin = WhoamiPlugin(self._instance)
        whoami_plugin.create()

class c001003006(baseconfig):
    def __init__(self, instance):
        super(c001003006, self).__init__(instance)
        self._operations = [
            # Create plugin configs first
            # c001003006_whoami_plugin(self._instance)
            # Create our sample entries.
            # op001003006_sample_entries(self._instance),
        ]


    # We need the following to be created and ENABLED
    # SyntaxValidationPlugin
    # SchemaReloadPlugin
    # StateChangePlugin
    # RolesPlugin
    # ACLPlugin
    # ACLPreoperationPlugin
    # ClassOfServicePlugin
    # ViewsPlugin
    # SevenBitCheckPlugin
    # AccountUsabilityPlugin
    # AutoMembershipPlugin
    # DereferencePlugin
    # HTTPClientPlugin
    # LinkedAttributesPlugin
    # ManagedEntriesPlugin
    # WhoamiPlugin
    # LDBMBackendPlugin
    # ChainingBackendPlugin
    




# --- BEGIN COPYRIGHT BLOCK ---
# Copyright (C) 2023 Red Hat, Inc.
# All rights reserved.
#
# License: GPL (version 3 or any later version).
# See LICENSE for details.
# --- END COPYRIGHT BLOCK ---

import ldap
import json
from lib389.plugins import (AutoMembershipPlugin, AutoMembershipDefinition, AutoMembershipDefinitions,
                            AutoMembershipRegexRule, AutoMembershipRegexRules, AutoMembershipFixupTasks)
from lib389.cli_conf import add_generic_plugin_parsers, generic_object_edit, generic_object_add
from lib389.cli_base import CustomHelpFormatter
from lib389.utils import get_task_status

arg_to_attr_definition = {
    'default_group': 'autoMemberDefaultGroup',
    'filter': 'autoMemberFilter',
    'grouping_attr': 'autoMemberGroupingAttr',
    'scope': 'autoMemberScope'
}

arg_to_attr_regex = {
    'exclusive': 'autoMemberExclusiveRegex',
    'inclusive': 'autoMemberInclusiveRegex',
    'target_group': 'autoMemberTargetGroup'
}


def definition_list(inst, basedn, log, args):
    automembers = AutoMembershipDefinitions(inst)
    result = []
    result_json = []
    for definition in automembers.list():
        if args.json:
            result_json.append(json.loads(definition.get_all_attrs_json()))
        else:
            result.append(definition.rdn)
    if args.json:
        log.info(json.dumps({"type": "list", "items": result_json}, indent=4))
    else:
        if len(result) > 0:
            for i in result:
                log.info(i)
        else:
            log.info("No Automember definitions were found")


def definition_add(inst, basedn, log, args):
    log = log.getChild('definition_add')
    plugin = AutoMembershipPlugin(inst)
    props = {'cn': args.DEFNAME}
    generic_object_add(AutoMembershipDefinition, inst, log, args, arg_to_attr_definition, basedn=plugin.dn, props=props)


def definition_edit(inst, basedn, log, args):
    log = log.getChild('definition_edit')
    definitions = AutoMembershipDefinitions(inst)
    definition = definitions.get(args.DEFNAME)
    generic_object_edit(definition, log, args, arg_to_attr_definition)


def definition_show(inst, basedn, log, args):
    log = log.getChild('definition_show')
    definitions = AutoMembershipDefinitions(inst)
    definition = definitions.get(args.DEFNAME)

    if not definition.exists():
        raise ldap.NO_SUCH_OBJECT("Entry %s doesn't exists" % args.DEFNAME)
    if args and args.json:
        o_str = definition.get_all_attrs_json()
        log.info(o_str)
    else:
        log.info(definition.display())


def definition_del(inst, basedn, log, args):
    # First delete regexes, then the defintion
    log = log.getChild('definition_del')
    definitions = AutoMembershipDefinitions(inst)
    definition = definitions.get(args.DEFNAME)
    regexes = AutoMembershipRegexRules(inst, definition.dn)
    for regex in regexes.list():
        regex.delete()
    definition.delete()
    log.info("Successfully deleted the %s definition", args.DEFNAME)


def regex_list(inst, basedn, log, args):
    definitions = AutoMembershipDefinitions(inst)
    definition = definitions.get(args.DEFNAME)
    regexes = AutoMembershipRegexRules(inst, definition.dn)
    result = []
    result_json = []
    for regex in regexes.list():
        if args.json:
            result_json.append(json.loads(regex.get_all_attrs_json()))
        else:
            result.append(regex.rdn)
    if args.json:
        log.info(json.dumps({"type": "list", "items": result_json}, indent=4))
    else:
        if len(result) > 0:
            for i in result:
                log.info(i)
        else:
            log.info("No Automember regexes were found")


def regex_add(inst, basedn, log, args):
    log = log.getChild('regex_add')
    definitions = AutoMembershipDefinitions(inst)
    definition = definitions.get(args.DEFNAME)
    props = {'cn': args.REGEXNAME}
    generic_object_add(AutoMembershipRegexRule, inst, log, args, arg_to_attr_regex, basedn=definition.dn, props=props)


def regex_edit(inst, basedn, log, args):
    log = log.getChild('regex_edit')
    definitions = AutoMembershipDefinitions(inst)
    definition = definitions.get(args.DEFNAME)
    regexes = AutoMembershipRegexRules(inst, definition.dn)
    regex = regexes.get(args.REGEXNAME)
    generic_object_edit(regex, log, args, arg_to_attr_regex)


def regex_show(inst, basedn, log, args):
    log = log.getChild('regex_show')
    definitions = AutoMembershipDefinitions(inst)
    definition = definitions.get(args.DEFNAME)
    regexes = AutoMembershipRegexRules(inst, definition.dn)
    regex = regexes.get(args.REGEXNAME)

    if not regex.exists():
        raise ldap.NO_SUCH_OBJECT("Entry %s doesn't exists" % args.REGEXNAME)
    if args and args.json:
        o_str = regex.get_all_attrs_json()
        log.info(o_str)
    else:
        log.info(regex.display())


def regex_del(inst, basedn, log, args):
    log = log.getChild('regex_del')
    definitions = AutoMembershipDefinitions(inst)
    definition = definitions.get(args.DEFNAME)
    regexes = AutoMembershipRegexRules(inst, definition.dn)
    regex = regexes.get(args.REGEXNAME)
    regex.delete()
    log.info("Successfully deleted the %s regex", args.REGEXNAME)


def fixup(inst, basedn, log, args):
    plugin = AutoMembershipPlugin(inst)
    log.info('Attempting to add task entry... This will fail if Automembership plug-in is not enabled.')
    if not plugin.status():
        log.error("'%s' is disabled. Rebuild membership task can't be executed" % plugin.rdn)
    fixup_task = plugin.fixup(args.DN, args.filter, args.cleanup)
    if args.wait:
        log.info(f'Waiting for fixup task "{fixup_task.dn}" to complete.  You can safely exit by pressing Control C ...')
        fixup_task.wait(timeout=args.timeout)
        exitcode = fixup_task.get_exit_code()
        if exitcode != 0:
            if excit_Code is None:
                raise ValueError(f'Rebuild membership task "{fixup_task.dn}" for {args.DN} has not completed. Please, check logs')
            else:
                raise ValueError(f'Rebuild membership task "{fixup_task.dn}" for {args.DN} has failed (error {exitcode}). Please, check logs')
        else:
            raise ValueError('Fixup task successfully completed')
    else:
        log.info(f'Successfully added task entry "{fixup_task.dn}". This task is running in the background. To track its progress you can use the "fixup-status" command.')


def do_fixup_status(inst, basedn, log, args):
    get_task_status(inst, log, AutoMembershipFixupTasks, dn=args.dn, show_log=args.show_log,
                    watch=args.watch, use_json=args.json)


def abort(inst, basedn, log, args):
    plugin = AutoMembershipPlugin(inst)
    log.info('Attempting to add abort task entry... This will fail if Automembership plug-in is not enabled.')
    if not plugin.status():
        log.error("'%s' is disabled. Abort rebuild membership task can't be executed" % plugin.rdn)
    fixup_task = plugin.abort_fixup(args.DN, args.filter)
    fixup_task.wait(timeout=args.timeout)
    exitcode = fixup_task.get_exit_code()
    if exitcode != 0:
        if exitcode is None:
            raise ValueError('Abort rebuild membership task has not completed. Please, check logs')
        else:
            raise ValueError('Abort rebuild membership task has failed. Please, check logs')
    else:
        log.info('Successfully added abort task entry')


def _add_parser_args_definition(parser):
    parser.add_argument('--grouping-attr',  required=True,
                        help='Specifies the name of the member attribute in the group entry and '
                             'the attribute in the object entry that supplies the member attribute value, '
                             'in the format group_member_attr:entry_attr (autoMemberGroupingAttr)')
    parser.add_argument('--default-group',
                        help='Sets default or fallback group to add the entry to as a member '
                             'attribute in group entry (autoMemberDefaultGroup)')
    parser.add_argument('--scope', required=True,
                        help='Sets the subtree DN to search for entries (autoMemberScope)')
    parser.add_argument('--filter', required=True,
                        help='Sets a standard LDAP search filter to use to search for '
                             'matching entries (autoMemberFilter)')


def _add_parser_args_regex(parser):
    parser.add_argument("--exclusive", nargs='+',
                        help='Sets a single regular expression to use to identify '
                             'entries to exclude (autoMemberExclusiveRegex)')
    parser.add_argument('--inclusive', nargs='+',
                        help='Sets a single regular expression to use to identify '
                             'entries to include (autoMemberInclusiveRegex)')
    parser.add_argument('--target-group', required=True,
                        help='Sets which group to add the entry to as a member, if it meets '
                             'the regular expression conditions (autoMemberTargetGroup)')


def create_parser(subparsers):
    automember = subparsers.add_parser('automember', help="Manage and configure Automembership plugin", formatter_class=CustomHelpFormatter)
    subcommands = automember.add_subparsers(help='action')
    add_generic_plugin_parsers(subcommands, AutoMembershipPlugin)

    automember_list = subcommands.add_parser('list', help='List Automembership definitions or regex rules.', formatter_class=CustomHelpFormatter)
    subcommands_list = automember_list.add_subparsers(help='action')
    list_definitions = subcommands_list.add_parser('definitions', help='Lists Automembership definitions.', formatter_class=CustomHelpFormatter)
    list_definitions.set_defaults(func=definition_list)
    list_regexes = subcommands_list.add_parser('regexes', help='List Automembership regex rules.', formatter_class=CustomHelpFormatter)
    list_regexes.add_argument('DEFNAME', help='The definition entry CN')
    list_regexes.set_defaults(func=regex_list)

    definition = subcommands.add_parser('definition', help='Manage Automembership definition.', formatter_class=CustomHelpFormatter)
    definition.add_argument('DEFNAME', help='The definition entry CN.')
    subcommands_definition = definition.add_subparsers(help='action')

    add_def = subcommands_definition.add_parser('add', help='Creates Automembership definition.', formatter_class=CustomHelpFormatter)
    add_def.set_defaults(func=definition_add)
    _add_parser_args_definition(add_def)
    edit_def = subcommands_definition.add_parser('set', help='Edits Automembership definition.', formatter_class=CustomHelpFormatter)
    edit_def.set_defaults(func=definition_edit)
    _add_parser_args_definition(edit_def)
    delete_def = subcommands_definition.add_parser('delete', help='Removes Automembership definition.', formatter_class=CustomHelpFormatter)
    delete_def.set_defaults(func=definition_del)
    show_def = subcommands_definition.add_parser('show', help='Displays Automembership definition.', formatter_class=CustomHelpFormatter)
    show_def.set_defaults(func=definition_show)

    regex = subcommands_definition.add_parser('regex', help='Manage Automembership regex rules.', formatter_class=CustomHelpFormatter)
    regex.add_argument('REGEXNAME', help='The regex entry CN')
    subcommands_regex = regex.add_subparsers(help='action')

    add_regex = subcommands_regex.add_parser('add', help='Creates Automembership regex.', formatter_class=CustomHelpFormatter)
    add_regex.set_defaults(func=regex_add)
    _add_parser_args_regex(add_regex)
    edit_regex = subcommands_regex.add_parser('set', help='Edits Automembership regex.', formatter_class=CustomHelpFormatter)
    edit_regex.set_defaults(func=regex_edit)
    _add_parser_args_regex(edit_regex)
    delete_regex = subcommands_regex.add_parser('delete', help='Removes Automembership regex.', formatter_class=CustomHelpFormatter)
    delete_regex.set_defaults(func=regex_del)
    show_regex = subcommands_regex.add_parser('show', help='Displays Automembership regex.', formatter_class=CustomHelpFormatter)
    show_regex.set_defaults(func=regex_show)

    fixup_task = subcommands.add_parser('fixup', help='Run a rebuild membership task.', formatter_class=CustomHelpFormatter)
    fixup_task.set_defaults(func=fixup)
    fixup_task.add_argument('DN', help="Base DN that contains entries to fix up")
    fixup_task.add_argument('-f', '--filter', required=True, help='Sets the LDAP filter for entries to fix up')
    fixup_task.add_argument('-s', '--scope', required=True, choices=['sub', 'base', 'one'], type=str.lower,
                            help='Sets the LDAP search scope for entries to fix up')
    fixup_task.add_argument('--cleanup', action='store_true',
                            help="Clean up previous group memberships before rebuilding")
    fixup_task.add_argument('--wait', action='store_true',
                            help="Wait for the task to finish, this could take a long time")
    fixup_task.add_argument('--timeout', default=0, type=int,
                            help="Set a timeout to wait for the fixup task.  Default is 0 (no timeout)")

    fixup_status = subcommands.add_parser('fixup-status', help='Check the status of a fix-up task', formatter_class=CustomHelpFormatter)
    fixup_status.set_defaults(func=do_fixup_status)
    fixup_status.add_argument('--dn', help="The task entry's DN")
    fixup_status.add_argument('--show-log', action='store_true', help="Display the task log")
    fixup_status.add_argument('--watch', action='store_true',
                              help="Watch the task's status and wait for it to finish")

    abort_fixup = subcommands.add_parser('abort-fixup', help='Abort the rebuild membership task.', formatter_class=CustomHelpFormatter)
    abort_fixup.set_defaults(func=abort)
    abort_fixup.add_argument('--timeout', default=0, type=int,
                            help="Set a timeout to wait for the abort task.  Default is 0 (no timeout)")